# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import builtins as bltn
import dataclasses as dcls
import datetime as dttm
import json
import pickle as pckl
from pathlib import Path as path_t
from typing import Any, Dict, Optional, Sequence, Union
from collections.abc import Collection, Mapping

import matplotlib.pyplot as pypl
import numpy as nmpy
import tensorflow.keras as kras
from mpl_toolkits.mplot3d import Axes3D as axes_3d_t


array_t = nmpy.ndarray


BUILTIN_TYPES = tuple(
    _typ for _elm in dir(bltn) if isinstance((_typ := getattr(bltn, _elm)), type)
)

# The following lists are meant to be safe enough, not to serve as references
PATH_ILLEGAL_CHARACTERS_LIN = r"/"
PATH_ILLEGAL_CHARACTERS_OSX = r":"
PATH_ILLEGAL_CHARACTERS_WIN = r'|/\<>:?*"'

REPLACEMENT_CHARACTER = "_"
VERSION_SEPARATOR = "-"

PATH_ILLEGAL_CHARACTERS = "".join(
    set(
        PATH_ILLEGAL_CHARACTERS_LIN
        + PATH_ILLEGAL_CHARACTERS_OSX
        + PATH_ILLEGAL_CHARACTERS_WIN
    )
)
if (REPLACEMENT_CHARACTER in PATH_ILLEGAL_CHARACTERS) or (
    VERSION_SEPARATOR in PATH_ILLEGAL_CHARACTERS
):
    raise ValueError(
        f'The character "{REPLACEMENT_CHARACTER}" or "{VERSION_SEPARATOR}" is an illegal path character'
    )


@dcls.dataclass(repr=False, eq=False)
class archiver_t:

    folder: path_t = None

    @classmethod
    def NewForFolderAndSequence(
        cls, folder: Union[str, path_t], sequence: Union[str, path_t], /
    ) -> archiver_t:
        """
        folder: If a path_t, then it must not contain any illegal characters for the target system. Otherwise, any
        illegal character will be replaced with a legal one (a priori, "_").
        sequence: Whether a str or a path_t, only the name (with extension) will be retained, the extension ".", if any,
        will be replaced with "_", and any illegal character will be replaced with a legal one (see "folder").
        """
        if isinstance(folder, str):
            folder = path_t(_ReplacePathIllegalCharacters(folder))
        if isinstance(sequence, path_t):
            sequence = _ReplacePathIllegalCharacters(
                sequence.stem + REPLACEMENT_CHARACTER + sequence.suffix[1:]
            )
        else:
            sequence = _ReplacePathIllegalCharacters(sequence)
        for component in (sequence, None):
            if folder.exists():
                if not folder.is_dir():
                    raise ValueError(
                        f"{folder}: Not a folder; Cannot be used by {cls.__name__.upper()}"
                    )
            else:
                folder.mkdir()
            if component is not None:
                folder /= component

        original_name = _ReplacePathIllegalCharacters(_TimeStamp())
        folder /= original_name
        version = 0
        while folder.exists():
            version += 1
            folder = folder.parent / f"{original_name}{VERSION_SEPARATOR}{version}"
        folder.mkdir()

        instance = cls(folder=folder)

        return instance

    def Store(
        self, element: Any, name: str, /, *, with_time_stamp: bool = True
    ) -> None:
        """
        name: Any illegal character will be replaced with a legal one (see "folder" in NewForFolderAndSequence)
        """
        should_log = False
        should_csv = False
        if name.lower().endswith(".log"):
            should_log = isinstance(element, str)
            if not should_log:
                raise ValueError(
                    f"{type(element).__name__}: Invalid type for logging; Expected=str"
                )
        elif name.lower().endswith(".csv"):
            should_csv = isinstance(element, array_t) and (element.ndim == 2)
            if not should_csv:
                if isinstance(element, array_t):
                    raise ValueError(
                        f"{element.ndim}: Invalid number of dimensions for CSV output; Expected=2"
                    )
                else:
                    raise ValueError(
                        f"{type(element).__name__}: Invalid type for CSV output; Expected=numpy.ndarray"
                    )

        if with_time_stamp and not (should_log or should_csv):
            name += _TimeStamp()
        name = _ReplacePathIllegalCharacters(name)

        if should_log:
            with open(self.folder / name, "a") as writer:
                writer.write(_TimeStamp() + "\n")
                writer.write(element + "\n")
        elif should_csv:
            nmpy.savetxt(self.folder / name, element, delimiter=",", fmt="%f")
        elif _IsFullyBuiltin(element):
            with open(self.folder / f"{name}.json", "w") as writer:
                json.dump(element, writer, indent=4)
        elif isinstance(element, array_t):
            nmpy.savez_compressed(self.folder / f"{name}.npz", contents=element)
        elif isinstance(element, pypl.Figure):
            element.savefig(self.folder / f"{name}.png")
        elif isinstance(element, pypl.Axes) or isinstance(element, axes_3d_t):
            element.figure.savefig(self.folder / f"{name}.png")
        elif isinstance(element, kras.Model):
            element.save(self.folder / name)
        else:
            raise ValueError(
                f"{type(element).__name__}: Type of {element} unprocessable by {self.__class__.__name__.upper()}"
            )

    def StoreWithPickle(
        self,
        value: Union[Any, Sequence[Any], Dict[str, Any]],
        pickle_name: str,
        /,
        *,
        name: Union[str, Sequence[str]] = None,
        with_time_stamp: bool = True,
    ) -> None:
        """
        value: see PickleAndStoreObjects
        """
        if with_time_stamp:
            pickle_name += _TimeStamp()
        pickle_name = _ReplacePathIllegalCharacters(pickle_name)

        PickleAndStoreObjects(value, self.folder / pickle_name, name=name)


def PickleAndStoreObjects(
    value: Union[Any, Sequence[Any], Dict[str, Any]],
    where: Union[str, path_t],
    /,
    *,
    name: Union[str, Sequence[str]] = None,
) -> None:
    """
    value: Pass globals() to store all builtin-type objects
    """
    if isinstance(name, str):
        names = (name,)
        values = (value,)
    elif (name is None) and isinstance(value, dict):
        # value is normally globals()
        names = []
        values = []
        for _nme, _val in value.items():
            if isinstance(_nme, str) and _IsFullyBuiltin(_val):
                names.append(_nme)
                values.append(_val)
    elif (
        isinstance(name, Sequence)
        and ((n_names := name.__len__()) > 0)
        and all(isinstance(_elm, str) for _elm in name)
        and isinstance(value, Sequence)
        and (value.__len__() == n_names)
    ):
        names = name
        values = value
    else:
        raise ValueError(
            f"name type={type(name).__name__}/value type={type(value).__name__}: Invalid calling context. "
            f"Expected=str/Any or None/dict or Sequence[str]/Sequence[Any] w/ same length."
        )

    data = {_nme: _val for _nme, _val in zip(names, values)}
    with open(where, "wb") as accessor:
        pckl.dump(data, accessor)


def LoadPickledObjects(
    where: Union[str, path_t], /, *, context: Dict[str, Any] = None
) -> Optional[Dict[str, Any]]:
    """
    context: Use globals() typically
    """
    with open(where, "rb") as accessor:
        data = pckl.load(accessor)

    if context is None:
        return data

    for name, value in data.items():
        context[name] = value

    return None


def _TimeStamp() -> str:
    """"""
    return dttm.datetime.now().isoformat().replace(":", ".")


def _ReplacePathIllegalCharacters(
    string: str, /, *, replacement: str = REPLACEMENT_CHARACTER
) -> str:
    """"""
    translations = str.maketrans(
        PATH_ILLEGAL_CHARACTERS, PATH_ILLEGAL_CHARACTERS.__len__() * replacement
    )
    output = string.translate(translations)

    return output


def _IsFullyBuiltin(element: Any) -> bool:
    """"""
    if type(element) in BUILTIN_TYPES:
        if isinstance(element, Mapping):
            return all(
                _IsFullyBuiltin(_key) and _IsFullyBuiltin(_val)
                for _key, _val in element.items()
            )
        elif isinstance(element, Collection):
            return all(_IsFullyBuiltin(_elm) for _elm in element)

        return True

    return False



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import numpy as nmpy
from cell_tracking_BC.type.frame import frame_t
from cell_tracking_BC.type.segmentation import segmentation_t


array_t = nmpy.ndarray


shape = (100, 100)

nucleus_map = nmpy.zeros(shape, dtype=nmpy.bool_)
nucleus_map[50:58, 60:72] = True

cytoplasm_map = nmpy.zeros(shape, dtype=nmpy.bool_)
cytoplasm_map[45:65, 54:80] = True
cytoplasm_map[nucleus_map] = False

frame = frame_t(nucleus_map + cytoplasm_map)
segmentation = segmentation_t.NewFromCompartments(
    nucleus=nucleus_map, cytoplasm=cytoplasm_map
)
frame.AddCellsFromSegmentation(segmentation)
print(frame)



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import NamedTuple as named_tuple_t
from typing import Type, TypeVar

# Can import everything but drawers and viewers
from cell_tracking_BC.in_out.graphics.type.axes import axes_2d_t, axes_3d_t
from cell_tracking_BC.in_out.graphics.type.figure import figure_t
from cell_tracking_BC.in_out.graphics.type.signatures import signatures_p


path_collection_h = TypeVar("path_collection_h")
# Do not import *_t (=> circular imports)
s_drawer_2d_h = TypeVar("s_drawer_2d_h")
s_viewer_2d_h = TypeVar("s_viewer_2d_h")
t_viewer_2d_h = TypeVar("t_viewer_2d_h")


class context_t(named_tuple_t):
    figure_2d_t: Type[figure_t] = None
    figure_3d_t: Type[figure_t] = None
    axes_2d_t: Type[axes_2d_t] = None
    axes_3d_t: Type[axes_3d_t] = None
    s_drawer_2d_t: Type[s_drawer_2d_h] = None
    s_viewer_2d_t: Type[s_viewer_2d_h] = None
    t_viewer_2d_t: Type[t_viewer_2d_h] = None
    #
    CellAnnotationStyle: signatures_p.cell_annotation_style_h = lambda: {}
    #
    NewSlider: signatures_p.new_slider_h = None
    SliderValue: signatures_p.slider_value_h = None
    UpdateSlider: signatures_p.update_slider_h = None
    SliderBounds: signatures_p.slider_bounds_h = None
    SliderAxes: signatures_p.slider_axes_h = lambda: None
    #
    IsVisible: signatures_p.get_visibility_h = None
    SetVisibility: signatures_p.set_visibility_h = None
    #
    IsTargetOfEvent: signatures_p.is_target_of_event_h = lambda: True
    KeyEventKey: signatures_p.key_event_key_h = None
    ScrollEventStep: signatures_p.scroll_event_step_h = None



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Optional, Sequence, Union

import numpy as nmpy

import cell_tracking_BC.in_out.graphics.generic.d_2 as gnr2
import cell_tracking_BC.in_out.graphics.generic.d_3 as gnr3
import cell_tracking_BC.in_out.graphics.generic.d_any as gnrc
from cell_tracking_BC.in_out.file.archiver import archiver_t
from cell_tracking_BC.in_out.graphics.dbe.matplotlib.context import DRAWING_CONTEXT
from cell_tracking_BC.in_out.graphics.type.axes import axes_2d_t, axes_3d_t
from cell_tracking_BC.in_out.graphics.type.context import context_t
from cell_tracking_BC.in_out.graphics.type.figure import figure_t
from cell_tracking_BC.in_out.graphics.type.s_viewer_2d import s_viewer_2d_t
from cell_tracking_BC.type.sequence import sequence_t


array_t = nmpy.ndarray


def ShowSequenceStatistics(
    sequence: sequence_t,
    /,
    *,
    channel: Union[str, Sequence[str]] = None,
    prepare_only: bool = False,
    interactively: bool = True,
    in_main_thread: bool = True,
    figure_name: str = "sequence-statistics",
    dbe: context_t = DRAWING_CONTEXT,
    archiver: archiver_t = None,
) -> Optional[figure_t]:
    """"""
    output = None

    if channel is None:
        channels = sequence.base_channels
    elif isinstance(channel, str):
        channels = (channel,)
    else:
        channels = channel

    statistics = ("amin", "amax", "mean", "median")
    ComputedStatistics = tuple(getattr(nmpy, _stt) for _stt in statistics)
    records = {_stt: {_chl: [] for _chl in channels} for _stt in statistics}
    for frames in sequence.Frames(channel=channels):
        for channel, frame in zip(channels, frames):
            for name, Computed in zip(statistics, ComputedStatistics):
                records[name][channel].append(Computed(frame))

    figure, all_axes = dbe.figure_2d_t.NewFigureAndAxes(n_rows=statistics.__len__())
    all_axes: Sequence[axes_2d_t]

    for s_idx, (name, values_per_channel) in enumerate(records.items()):
        for channel, values in values_per_channel.items():
            all_axes[s_idx].PlotLines(range(values.__len__()), values, label=channel)

    for name, axes in zip(records.keys(), all_axes):
        axes.SetTimeAxisProperties(sequence.length - 1)
        axes.ShowLegend()
        axes.SetTitle(name)

    figure.Archive(name=figure_name, archiver=archiver)
    if prepare_only:
        output = figure
    else:
        figure.Show(interactively=interactively, in_main_thread=in_main_thread)

    return output


def ShowSequence(
    sequence: Union[Sequence[array_t], sequence_t],
    /,
    *,
    channel: str = None,
    with_segmentation: bool = True,
    with_cell_labels: bool = True,
    with_track_labels: bool = True,
    mode: str = "2d+t",
    keep_every: int = 2,
    n_levels: int = 100,
    iso_value: float = None,
    with_ticks: bool = True,
    with_colorbar: bool = True,
    prepare_only: bool = False,
    interactively: bool = True,
    in_main_thread: bool = True,
    dbe: context_t = DRAWING_CONTEXT,
) -> Optional[Union[figure_t, s_viewer_2d_t]]:
    """
    mode: "2d+t", "mille-feuille", "tunnels"
    """
    output = None

    if mode == "2d+t":
        viewer = dbe.s_viewer_2d_t.NewForChannels(
            sequence,
            dbe,
            channel=channel,
            with_segmentation=with_segmentation,
            with_cell_labels=with_cell_labels,
            with_track_labels=with_track_labels,
            with_ticks=with_ticks,
            with_colorbar=with_colorbar,
        )
        viewer.SetConversionToAnnotatedVolume(gnr2.AsAnnotatedVolume)
        if prepare_only:
            output = viewer
        else:
            viewer.Show(interactively=interactively, in_main_thread=in_main_thread)
    elif mode in ("mille-feuille", "tunnels"):
        if isinstance(sequence, sequence_t):
            if channel is None:
                channel = sequence.cell_channel
            frames = sequence.Frames(channel=channel)
        else:
            frames = sequence

        figure, axes = dbe.figure_3d_t.NewFigureAndAxes()
        axes: axes_3d_t
        if mode == "mille-feuille":
            if isinstance(sequence, sequence_t) and sequence.has_cells:
                cell_frames = sequence.cell_frames
            else:
                cell_frames = None
            cell_contours = gnrc.CellContours(sequence, with_segmentation)
            tracks = gnrc.CellTracks(sequence, with_track_labels)
            gnr3.ShowFramesAsMilleFeuille(
                frames,
                cell_contours,
                with_cell_labels,
                cell_frames,
                tracks,
                axes,
                dbe.CellAnnotationStyle,
                keep_every=keep_every,
                n_levels=n_levels,
            )
        else:
            if iso_value is None:
                iso_value = nmpy.percentile(frames[0], 75).item()
            gnr3.ShowFramesAsTunnels(
                frames, axes, keep_every=keep_every, iso_value=iso_value
            )

        if prepare_only:
            output = figure
        else:
            figure.Show(interactively=interactively, in_main_thread=in_main_thread)
    else:
        raise ValueError(
            f'{mode}: Invalid mode; Expected="2d+t", "mille-feuille", "tunnels"'
        )

    return output



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from multiprocessing import Process as process_t
from typing import Sequence, Tuple, Union

try:
    # noinspection PyPackageRequirements
    from vedo import Plotter as vedo_figure_t
except ModuleNotFoundError:
    vedo_figure_t = None

import cell_tracking_BC.in_out.graphics.dbe.vedo.missing as vmsg
from cell_tracking_BC.in_out.graphics.type.figure import figure_t as base_figure_t


_UNSHOWN_FIGURES = []
_BBOX_AXES_MARGIN = 0.05


if vedo_figure_t is None:
    figure_t = vmsg.RaiseMissingVedoException
else:
    @dtcl.dataclass(init=False, repr=False, eq=False)
    class figure_t(vedo_figure_t, base_figure_t):

        bbox: Sequence[Sequence[float]] = None

        @classmethod
        def NewFigureAndAxes(
            cls,
            /,
            *,
            n_rows: int = 1,
            n_cols: int = 1,
            title: str = None,
            offline_version: bool = False,
        ) -> Tuple[
            figure_t,
            Union[figure_t, Sequence[figure_t], Sequence[Sequence[figure_t]]],
        ]:
            """"""
            figure = cls()
            axes = figure

            _UNSHOWN_FIGURES.append(figure)

            return figure, axes

        def UpdateBBoxFromOne(self, new_point: Sequence[float], /) -> None:
            """"""
            if self.bbox is None:
                self.bbox = (new_point, new_point)
            else:
                minima = [min(_new, _old) for _new, _old in zip(new_point, self.bbox[0])]
                maxima = [max(_new, _old) for _new, _old in zip(new_point, self.bbox[1])]
                self.bbox = (minima, maxima)

        def UpdateBBoxFromMany(
            self, exc_s: Sequence[float], why_s: Sequence[float], zee_s: Sequence[float], /
        ) -> None:
            """"""
            minima = [min(exc_s), min(why_s), min(zee_s)]
            maxima = [max(exc_s), max(why_s), max(zee_s)]
            if self.bbox is not None:
                minima = [min(_new, _old) for _new, _old in zip(minima, self.bbox[0])]
                maxima = [max(_new, _old) for _new, _old in zip(maxima, self.bbox[1])]
            self.bbox = (minima, maxima)

        @property
        def bbox_extent(self) -> Sequence[float]:
            """"""
            return tuple(_max - _min for _min, _max in zip(*self.bbox))

        @property
        def bbox_origin_and_corner(self) -> Tuple[Sequence[float], Sequence[float]]:
            """"""
            extent = self.bbox_extent
            output = tuple(
                (_min - _BBOX_AXES_MARGIN * _xtt, _max + _BBOX_AXES_MARGIN * _xtt)
                for _min, _max, _xtt in zip(self.bbox[0], self.bbox[1], extent)
            )

            return tuple(zip(*output))

        def Show(
            self,
            /,
            *,
            interactively: bool = True,
            in_main_thread: bool = True,
        ) -> None:
            """"""
            _UNSHOWN_FIGURES.remove(self)
            if in_main_thread:
                self.show().close()
            else:
                # Do not use "target=self.show().close"
                thread = process_t(target=lambda: self.show().close())
                thread.start()

        @staticmethod
        def ShowAll(
            *,
            interactively: bool = True,
            in_main_thread: bool = True,
        ) -> None:
            """"""
            while _UNSHOWN_FIGURES.__len__() > 1:
                _UNSHOWN_FIGURES[0].Show(interactively=interactively, in_main_thread=False)
            if _UNSHOWN_FIGURES.__len__() > 0:
                _UNSHOWN_FIGURES[0].Show(
                    interactively=interactively, in_main_thread=in_main_thread
                )

        def Figure(self) -> figure_t:
            """"""
            return self



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
import datetime as dttm
import re as regx
from pathlib import Path as path_t
from typing import Callable, Dict, Sequence, Tuple, Union

import numpy as nmpy
import tifffile as tiff

from cell_tracking_BC.in_out.graphics.type.axes import axes_2d_t as axes_t
from cell_tracking_BC.in_out.graphics.type.context import context_t
from cell_tracking_BC.in_out.graphics.type.event import (
    event_h,
    event_key_h,
    event_scroll_h,
)
from cell_tracking_BC.in_out.graphics.type.s_drawer_2d import s_drawer_2d_t
from cell_tracking_BC.type.frame import frame_t
from cell_tracking_BC.type.sequence import AllChannelsOfSequence, sequence_h


array_t = nmpy.ndarray

all_versions_h = Dict[
    str, Tuple[Tuple[int, int], Union[Sequence[array_t], Sequence[frame_t]]]
]
conversion_fct_h = Callable[..., array_t]


@dtcl.dataclass(repr=False, eq=False)
class s_viewer_2d_t(s_drawer_2d_t):
    _AsAnnotatedVolume: conversion_fct_h = None

    @classmethod
    def _NewForSequence(
        cls,
        sequence: sequence_h,
        AllVersionsOfSequence: Callable[[sequence_h], Tuple[all_versions_h, str]],
        dbe: context_t,
        /,
        *,
        version: str = None,
        with_segmentation: bool = False,
        with_cell_labels: bool = True,
        cell_frames: Sequence[frame_t] = None,
        with_track_labels: bool = True,
        in_axes: axes_t = None,
        with_ticks: bool = True,
    ) -> s_viewer_2d_t:
        """"""
        drawer = s_drawer_2d_t._NewForSequence(
            sequence,
            AllVersionsOfSequence,
            dbe,
            version=version,
            with_segmentation=with_segmentation,
            with_cell_labels=with_cell_labels,
            cell_frames=cell_frames,
            with_track_labels=with_track_labels,
            in_axes=in_axes,
            with_ticks=with_ticks,
        )
        if in_axes is None:
            current_time_point = None
            slider = dbe.NewSlider(drawer.figure, sequence.length)
        else:
            current_time_point = 0
            slider = None

        arguments = {
            "figure": drawer.figure,
            "axes": drawer.axes,
            "all_versions": drawer.all_versions,
            "current_time_point": current_time_point,
            "slider": slider,
        }
        instance = cls(**arguments)
        already_set = tuple(arguments.keys())
        for field in dtcl.fields(drawer):
            if (name := field.name) not in already_set:
                setattr(instance, name, getattr(drawer, name))

        instance._ActivateEventProcessing(instance.all_versions.__len__() > 1)

        return instance

    def _ActivateEventProcessing(self, more_than_one_version: bool, /) -> None:
        """"""
        self.figure.ActivateEvent("key_press_event", self._OnKeyPress)
        if more_than_one_version:
            self.figure.ActivateEvent("button_press_event", self._OnButtonPress)
        if self.slider is not None:
            self.figure.ActivateEvent("scroll_event", self._OnScrollEvent)

    def ShowNextVersion(self, /, *, should_draw: bool = True) -> None:
        """"""
        all_names = tuple(self.all_versions.keys())
        where = all_names.index(self.current_version)
        where = (where + 1) % all_names.__len__()
        new_version = all_names[where]

        self.ShowFrame(
            version=new_version, should_update_limits=True, should_draw=should_draw
        )

    def Show(
        self,
        /,
        *,
        interactively: bool = True,
        in_main_thread: bool = True,
    ) -> None:
        """"""
        self.figure.Show(interactively=interactively, in_main_thread=in_main_thread)

    def _OnKeyPress(
        self,
        event: event_key_h,
        /,
    ) -> None:
        """"""
        if self.dbe.KeyEventKey(event).lower() == "s":
            if self._AsAnnotatedVolume is None:
                print(
                    f'"{self.__class__.SetConversionToAnnotatedVolume.__name__}" has not been called; '
                    f'Saving the sequence is therefore disabled'
                )
                return

            print("Sequence saving in progress...")
            volume = self.AsAnnotatedVolume()

            illegal = "[^-_a-zA-Z0-9]"
            version = regx.sub(illegal, "", self.current_version)
            now = regx.sub(illegal, "-", dttm.datetime.now().isoformat())
            path = path_t.home() / f"sequence-{version}-{now}.tif"
            if path.exists():
                print(f"{path}: Existing path; Cannot override")
                return

            tiff.imwrite(
                str(path),
                volume,
                photometric="rgb",
                compression="deflate",
                planarconfig="separate",
                metadata={"axes": "XYZCT"},
            )
            print(f"Annotated sequence saved at: {path}")

    def _OnButtonPress(
        self,
        event: event_h,
        /,
    ) -> None:
        """"""
        if self.dbe.IsTargetOfEvent(self.axes, event):
            self.ShowNextVersion()
        elif (self.slider is not None) and self.dbe.IsTargetOfEvent(
            self.dbe.SliderAxes(self.slider), event
        ):
            self.ShowFrame(
                time_point=self.dbe.SliderValue(self.slider), force_update=True
            )

    def _OnScrollEvent(self, event: event_scroll_h) -> None:
        """"""
        value = self.dbe.SliderValue(self.slider)
        bounds = self.dbe.SliderBounds(self.slider)
        new_value = round(value + nmpy.sign(self.dbe.ScrollEventStep(event)))
        new_value = min(max(new_value, bounds[0]), bounds[1])
        if new_value != value:
            self.ShowFrame(time_point=new_value)

    def SetConversionToAnnotatedVolume(
        self, ConversionFct: conversion_fct_h, /
    ) -> None:
        """
        The goal of this method is to provide a "solution" to circular imports
        """
        self._AsAnnotatedVolume = ConversionFct

    def AsAnnotatedVolume(self) -> array_t:
        """"""
        frames = self.all_versions[self.current_version][1]
        drawer = self.__class__._NewForSequence(
            frames,
            AllChannelsOfSequence,
            self.dbe,
            with_segmentation=self.cell_contours is not None,
            with_cell_labels=self.with_cell_labels,
            cell_frames=self.cell_frames,
            with_track_labels=self.tracks is not None,
            with_ticks=False,
        )
        drawer.cell_contours = self.cell_contours
        drawer.tracks = self.tracks

        return self._AsAnnotatedVolume(
            frames,
            "",
            False,
            False,
            False,
            drawer=drawer,
        )



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Any, Dict
from typing import Protocol, Sequence

# Can import everything
from cell_tracking_BC.in_out.graphics.type.axes import axes_2d_t, axes_t
from cell_tracking_BC.in_out.graphics.type.event import event_h, event_key_h, event_scroll_h
from cell_tracking_BC.in_out.graphics.type.figure import figure_t
from cell_tracking_BC.in_out.graphics.type.widget import element_h, slider_h


class signatures_p(Protocol):
    @staticmethod
    def cell_annotation_style_h(
        highlighted: bool, multi_track: bool, /
    ) -> Dict[str, Any]:
        ...

    @staticmethod
    def new_slider_h(figure: figure_t, n_steps: int, /) -> slider_h:
        ...

    @staticmethod
    def slider_value_h(slider: slider_h, /) -> float:
        ...

    @staticmethod
    def update_slider_h(slider: slider_h, value: float, /) -> None:
        ...

    @staticmethod
    def slider_bounds_h(slider: slider_h, /) -> Sequence[float]:
        ...

    @staticmethod
    def slider_axes_h(slider: slider_h, /) -> axes_2d_t:
        ...

    @staticmethod
    def get_visibility_h(what: element_h, /) -> bool:
        ...

    @staticmethod
    def set_visibility_h(what: element_h, visibility: bool, /) -> None:
        ...

    @staticmethod
    def is_target_of_event_h(axes: axes_t, event: event_h, /) -> bool:
        ...

    @staticmethod
    def key_event_key_h(event: event_key_h, /) -> str:
        ...

    @staticmethod
    def scroll_event_step_h(event: event_scroll_h, /) -> float:
        ...



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Dict, Sequence, Union

import numpy as nmpy
import scipy.ndimage.morphology as scph
import skimage.exposure as xpsr
import skimage.morphology as mrph
import skimage.registration as rgst
import skimage.transform as trsf

from cell_tracking_BC.type.frame import frame_t


array_t = nmpy.ndarray


def WithImprovedLocalContrast(
    frame: frame_t,
    /,
    *,
    kernel_size: Union[int, Sequence[int]] = None,
    clip_limit: float = 0.01,
    n_bins: int = 256,
) -> array_t:
    """"""
    return xpsr.equalize_adapthist(
        frame, kernel_size=kernel_size, clip_limit=clip_limit, nbins=n_bins
    )


def RegisteredInTranslation(
    unregistered: frame_t,
    /,
    *,
    channels: Dict[str, frame_t] = None,
    reference: str = None,
    should_use_precomputed: bool = False,
) -> array_t:
    """
    /!\ If should_use_precomputed is True, then be sure to apply the transforms in a channel order such that the channel
    used to compute (and use) the shift (as opposed to just using the previously computed shift) is transformed first.
    """
    reference_frame = channels[reference]
    if unregistered is reference_frame:
        return unregistered

    if should_use_precomputed and ("shift" in reference_frame.runtime):
        shift = reference_frame.runtime["shift"]
    else:
        # Obsolete call: skimage.feature.register_translation
        shift = rgst.phase_cross_correlation(
            reference_frame,
            unregistered,
            upsample_factor=8,
            return_error=False,
        )
        reference_frame.runtime["shift"] = shift
    if nmpy.any(shift != 0.0):
        translation = trsf.EuclideanTransform(translation=shift)
        output = trsf.warp(unregistered, translation, preserve_range=True)
    else:
        output = unregistered

    return output


def WithSmallObjectsAndHolesRemoved(
    frame: frame_t, min_object_area: int, max_hole_area: int, /
) -> array_t:
    """"""
    big_objects = mrph.remove_small_objects(frame, min_size=min_object_area)
    output = WithSmallHolesRemoved(big_objects, max_hole_area=max_hole_area)

    return output


def WithSmallHolesRemoved(frame: frame_t, /, *, max_hole_area: int = None) -> array_t:
    """"""
    output = scph.binary_fill_holes(frame)
    if max_hole_area is None:
        return output

    holes = output.copy()
    holes[frame] = 0
    labeled, n_holes = mrph.label(holes, return_num=True, connectivity=1)
    for label in range(1, n_holes + 1):
        where_label = holes == label
        area = nmpy.count_nonzero(where_label)
        if area <= max_hole_area:
            output[where_label] = 0

    return output



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Dict, Tuple

import networkx as grph
from cell_tracking_BC.type.cell import cell_t
from cell_tracking_BC.type.track import (
    forking_track_t,
    unstructured_track_t,
)


def VersionOfForkingForLayout(
    track: forking_track_t,
) -> Tuple[forking_track_t, Dict[int, cell_t]]:
    """"""
    integer_to_cell = {_key: _val for _key, _val in enumerate(track)}
    cell_to_integer = {_key: _val for _val, _key in enumerate(track)}

    with_int_labels = grph.relabel_nodes(track, cell_to_integer)
    unstructured = unstructured_track_t(with_int_labels)
    forking, _ = forking_track_t.NewFromUnstructuredTrack(unstructured, 0)

    max_length = max(forking.lengths)
    next_idx = 0
    for single in forking.single_tracks_iterator:
        if single.length < max_length:
            n_missing = max_length - single.length
            latest_node = single[-1]
            for index in range(next_idx, next_idx + n_missing):
                new_node = f"f{index}"
                forking.add_edge(latest_node, new_node)
                latest_node = new_node
            next_idx += n_missing

    return forking, integer_to_cell



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dtcl
from typing import Sequence, Tuple, Union

import numpy as nmpy

try:
    # noinspection PyPackageRequirements
    import vedo.colors as clrs
    # noinspection PyPackageRequirements
    from vedo import Axes as vedo_axes_t
    # noinspection PyPackageRequirements
    from vedo import Box as box_t
    # noinspection PyPackageRequirements
    from vedo import Picture as picture_t
    # noinspection PyPackageRequirements
    from vedo import Volume as volume_t
    # noinspection PyPackageRequirements
    from vedo.shapes import Lines as lines_t
    # noinspection PyPackageRequirements
    from vedo.shapes import Text3D as text_t
except ModuleNotFoundError:
    clrs = vedo_axes_t = box_t = picture_t = volume_t = lines_t = text_t = None

import cell_tracking_BC.in_out.graphics.dbe.vedo.missing as vmsg
import cell_tracking_BC.in_out.graphics.dbe.vedo.style as styl
from cell_tracking_BC.in_out.graphics.dbe.vedo.d_any import figure_t as figure_anyd_t
from cell_tracking_BC.in_out.graphics.type.annotation import (
    annotation_h,
    cell_annotation_h,
)
from cell_tracking_BC.in_out.graphics.type.axes import axes_3d_t as base_axes_3d_t
from cell_tracking_BC.in_out.graphics.type.color import (
    ColorAndAlpha,
    ZeroOneValueToRGBAWithMilestones,
    colormap_h,
)


array_t = nmpy.ndarray


_MAX_N_TIME_LABELS = 15


if clrs is None:
    figure_t = vmsg.RaiseMissingVedoException
else:
    @dtcl.dataclass(init=False, repr=False, eq=False)
    class figure_t(figure_anyd_t, base_axes_3d_t):
        def SetTimeAxisProperties(self, latest: int, /) -> None:
            """"""
            origin, corner = self.bbox_origin_and_corner

            length, width, height = (_max - _min for _min, _max in zip(origin, corner))

            x_range, y_range, z_range = ((_min, _max) for _min, _max in zip(origin, corner))
            n_time_labels = min(latest + 1, _MAX_N_TIME_LABELS)
            time_labels = tuple(
                (_pos, _lbl)
                for _pos, _lbl in zip(
                    nmpy.linspace(self.bbox[0][2], self.bbox[1][2], num=n_time_labels),
                    nmpy.around(nmpy.linspace(0, latest, num=n_time_labels), decimals=1),
                )
            )

            box = box_t(pos=origin, length=length, width=width, height=height).alpha(0.1)
            axes = vedo_axes_t(
                box,
                xrange=x_range,
                yrange=y_range,
                zrange=z_range,
                zValuesAndLabels=time_labels,
                c="black",
            )

            self.__iadd__(axes)

        @staticmethod
        def TimeScaling(shape: Sequence[int], length: int, /) -> float:
            """"""
            size = 0.5 * sum(shape)

            return 2.0 * size / (length - 1)

        @staticmethod
        def MillefeuilleScaling(shape: Sequence[int], length: int, /) -> float:
            """"""
            size = 0.5 * sum(shape)

            return 0.1 * size

        def AddColormapFromMilestones(
            self,
            name: str,
            milestones: Sequence[Tuple[float, str]],
            /,
            *,
            position: str = "right",
        ) -> colormap_h:
            """"""
            # Leave it in 3-D (as opposed to any-D) because of 3-D box below
            output = lambda _vle: ZeroOneValueToRGBAWithMilestones(
                _vle, milestones, clrs.getColor
            )

            discretized = tuple(
                output(_vle)[:3] for _vle in nmpy.linspace(0.0, 1.0, num=100)
            )
            box = box_t(pos=(0, 0, 0), length=10, width=10, height=10).alpha(0.0)
            box.cmap(discretized, box.points()[:, 2])
            box.addScalarBar(title=name, nlabels=0)

            self.__iadd__(box)

            return output

        def PlotLines(self, xs, ys, *args, zdir="z", **kwargs):
            """"""
            zs = args[0]

            if "color" in kwargs:
                color, alpha = ColorAndAlpha(kwargs["color"], clrs.getColor)
            else:
                color, alpha = "black", None

            points = list(zip(ys, xs, zs))
            lines = lines_t(points[:-1], points[1:], c=color, lw=2)
            if alpha is not None:
                lines.alpha(alpha)

            self.__iadd__(lines)
            self.UpdateBBoxFromMany(ys, xs, zs)

        def PlotText(self, x, y, z, s, zdir=None, **kwargs):
            """"""
            position = [y, x, z]
            additionals = styl.ConvertedTextStyle(kwargs)
            if "alpha" in additionals:
                alpha = additionals["alpha"]
                del additionals["alpha"]
            else:
                alpha = None

            text = text_t(s, position, **additionals)
            if alpha is not None:
                text.alpha(alpha)

            text.followCamera()

            self.__iadd__(text)
            self.UpdateBBoxFromOne(position)

        def PlotCellAnnotation(
            self, position: Tuple[float, float, float], text: cell_annotation_h, /, **kwargs
        ) -> Union[annotation_h, Sequence[annotation_h]]:
            """"""
            additionals = kwargs.copy()
            if "alpha" in additionals:
                alpha = additionals["alpha"]
                del additionals["alpha"]
            else:
                alpha = None

            # TODO: multi-piece text is currently dealt with as single-piece text
            if not isinstance(text, str):
                text = "".join(_elm[0] for _elm in text)

            flipped = [position[1], position[0], position[2]]
            text = text_t(text, flipped, **additionals)
            if alpha is not None:
                text.alpha(alpha)
            text.followCamera()

            self.__iadd__(text)
            self.UpdateBBoxFromOne(flipped)

            return None

        def PlotImageInZ(
            self,
            image: array_t,
            all_rows: array_t,
            all_cols: array_t,
            n_levels: int,
            height: float,
            z_scaling: float,
            /,
            min_intensity: float = 0.0,
            intensity_range: float = 1.0,
            alpha: float = 0.8,
            **kwargs,
        ) -> None:
            """"""
            scaled = 255.0 * (image - min_intensity) / intensity_range
            picture = picture_t(scaled).z(z_scaling * height).alpha(alpha)

            self.__iadd__(picture)
            self.UpdateBBoxFromMany(all_cols.flatten(), all_rows.flatten(), (height,))

        def PlotIsosurface(
            self,
            volume: array_t,
            iso_value: float,
            /,
            should_be_capped: bool = False,
            keep_every: int = 2,
            **kwargs,
        ) -> None:
            """"""
            if should_be_capped:
                cap = nmpy.zeros_like(volume[0, ...])[None, ...]
                volume = nmpy.concatenate((cap, volume, cap), axis=0)

            isosurface = volume_t(volume).isosurface(threshold=[iso_value])
            time_scaling = self.__class__.TimeScaling(volume.shape[:2], volume.shape[2])
            isosurface.scale(s=(1.0, 1.0, time_scaling))

            self.__iadd__(isosurface)
            self.UpdateBBoxFromMany(
                isosurface.xbounds(), isosurface.ybounds(), isosurface.zbounds()
            )


axes_t = figure_t



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Callable, Optional, Sequence, Tuple, Union


rgb_color_h = Tuple[float, float, float]
rgba_color_h = Tuple[float, float, float, float]
colormap_h = Callable[
    [Union[float, Sequence[float]]], Union[rgba_color_h, Sequence[rgba_color_h]]
]


def ColorAndAlpha(
    color: Union[str, rgb_color_h, rgba_color_h],
    NameToRGB: Callable[[str], rgb_color_h],
    /,
    *,
    convert_to_rgb: bool = False,
) -> Tuple[Union[str, rgb_color_h], Optional[float]]:
    """"""
    if (is_str := isinstance(color, str)) or (color.__len__() == 3):
        alpha = None
        if is_str and convert_to_rgb:
            color = NameToRGB(color)
    else:
        alpha = color[-1]
        color = color[:-1]

    return color, alpha


def ZeroOneValueToRGBAWithMilestones(
    value: Union[float, Sequence[float]],
    milestones: Sequence[Tuple[float, str]],
    NameToRGB: Callable[[str], rgb_color_h],
    /,
) -> Union[rgba_color_h, Sequence[rgba_color_h]]:
    """"""
    if isinstance(value, Sequence):
        return tuple(
            ZeroOneValueToRGBAWithMilestones(_vle, milestones, NameToRGB)
            for _vle in value
        )

    n_milestones = milestones.__len__()
    m_idx = 0
    while (m_idx < n_milestones) and (value > milestones[m_idx][0]):
        m_idx += 1
    if m_idx >= n_milestones:
        color = milestones[-1][1]
    elif value < milestones[m_idx][0]:
        if m_idx > 0:
            previous = m_idx - 1
        else:
            previous = 0
        interval = (milestones[previous][1], milestones[m_idx][1])
        interval = (NameToRGB(_clr) for _clr in interval)
        ratio = (value - milestones[previous][0]) / (
            milestones[m_idx][0] - milestones[previous][0]
        )
        color = tuple(
            ratio * _end + (1.0 - ratio) * _stt for _stt, _end in zip(*interval)
        )
    else:
        color = milestones[m_idx][1]

    if isinstance(color, str):
        color = NameToRGB(color)
    color = (*color, 1.0)

    return color



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dtcl
from typing import Any, Optional, Sequence, Tuple, TypeVar, Union

import numpy as nmpy
import skimage.measure as msre

# Caution to avoid circular imports
import cell_tracking_BC.standard.uid as uqid
from cell_tracking_BC.in_out.graphics.type.annotation import (
    annotation_h,
    cell_annotation_h,
)
from cell_tracking_BC.in_out.graphics.type.color import colormap_h
from cell_tracking_BC.type.cell import cell_t
from cell_tracking_BC.type.frame import frame_t
from cell_tracking_BC.type.tracks import tracks_t


# 2 lines below: Avoid circular imports
cell_annotation_style_h = TypeVar("cell_annotation_style_h")
figure_h = TypeVar("figure_h")
#
array_t = nmpy.ndarray


@dtcl.dataclass(init=False, repr=False, eq=False)
class axes_t:
    def SetTitle(self, title: str, /) -> None:
        pass

    def SetAxisTicks(
        self,
        which: Union[str, Any],
        positions: Sequence[float],
        labels: Sequence[str],
        /,
    ) -> None:
        pass

    def SetTimeAxisProperties(self, latest: int, /) -> None:
        pass

    def TurnTicksOff(self) -> None:
        pass

    def PlotLines(
        self, xs, ys, *args, zdir="z", scalex=True, scaley=True, data=None, **kwargs
    ):
        pass

    def PlotCellAnnotation(
        self, position: Sequence[float], text: cell_annotation_h, /, **kwargs
    ) -> Union[annotation_h, Sequence[annotation_h]]:
        pass

    def PlotCellsDetails(
        self,
        frame: Union[array_t, frame_t],
        cell_contours: Optional[Sequence[array_t]],
        with_cell_labels: bool,
        cell_frame: Optional[frame_t],
        tracks: Optional[tracks_t],
        AnnotationStyle: cell_annotation_style_h,
        /,
        *,
        highlighted: int = -1,
        elevation: float = None,
        with_alpha_cell_uids: bool = True,
    ) -> Sequence[Tuple[int, Union[annotation_h, Sequence[annotation_h]]]]:
        """"""
        output = []

        if with_cell_labels or (tracks is not None):
            if elevation is None:
                self.RemoveLinesAndAnnotations()

            if cell_frame is None:
                labeled = msre.label(frame, connectivity=1)
                cells = msre.regionprops(labeled)
            else:
                cells = cell_frame.cells
            assert hasattr(cells[0], "centroid") and hasattr(
                cells[0], "label"
            ), "Please contact developer about API change"

            for cell in cells:
                if elevation is None:
                    position = nmpy.flipud(cell.centroid)
                else:
                    position = (*cell.centroid, elevation)

                text = []
                if with_cell_labels:
                    text.append(uqid.AlphaColumnFromLabel(cell.label))
                else:
                    text.append("")
                if (tracks is None) or not isinstance(cell, cell_t):
                    text.append("")
                else:
                    labels = tracks.TrackLabelsContainingCell(cell, tolerant_mode=True)
                    if labels is None:
                        text.append("?")
                    else:
                        if labels.__len__() > 1:
                            labels = "\n".join(str(_lbl) for _lbl in labels)
                        else:
                            labels = str(labels[0])
                        text.append(labels)

                if with_alpha_cell_uids:
                    text = "".join(text)
                    additionals = AnnotationStyle(
                        cell.label == highlighted, "\n" in text
                    )
                else:
                    if text[0].__len__() > 0:
                        multi_text = [(text[0], {})]
                    else:
                        multi_text = []
                    if text[1].__len__() > 0:
                        multi_text.extend(
                            (_pce, {"rotation": -90.0}) for _pce in text[1].split("\n")
                        )
                    text = multi_text
                    additionals = AnnotationStyle(cell.label == highlighted, False)

                annotation = self.PlotCellAnnotation(position, text, **additionals)

                output.append((cell.label, annotation))

        # Leave this block after cell annotation since, if placed before, the (new) contours are considered as previous
        # artists and removed.
        if cell_contours is not None:
            if elevation is None:
                for contour in cell_contours:
                    self.PlotLines(
                        contour[:, 1],
                        contour[:, 0],
                        linestyle=":",
                        color=(0.0, 1.0, 1.0, 0.3),
                    )
            else:
                for contour in cell_contours:
                    heights = contour.shape[0] * [elevation]
                    self.PlotLines(
                        contour[:, 0],
                        contour[:, 1],
                        heights,
                        linestyle=":",
                        color=(0.0, 1.0, 1.0, 0.3),
                    )

        return output

    def ShowLegend(self) -> None:
        pass

    def AddStandardColormap(
        self, name: str, colormap: str, /, *, position: str = "right"
    ) -> colormap_h:
        pass

    def AddColormapFromMilestones(
        self,
        name: str,
        milestones: Sequence[Tuple[float, str]],
        /,
        *,
        position: str = "right",
    ) -> colormap_h:
        pass

    def Freeze(self) -> None:
        pass

    def RemoveLinesAndAnnotations(self) -> None:
        pass

    def Figure(self) -> figure_h:
        pass


@dtcl.dataclass(init=False, repr=False, eq=False)
class axes_2d_t(axes_t):
    def SetTrackingAxesProperties(self, tick_positions: Sequence[float], /) -> None:
        pass

    def PlotPoints(
        self,
        x,
        y,
        s=None,
        c=None,
        marker=None,
        cmap=None,
        norm=None,
        vmin=None,
        vmax=None,
        alpha=None,
        linewidths=None,
        *,
        edgecolors=None,
        plotnonfinite=False,
        data=None,
        **kwargs,
    ):
        pass

    def PlotImage(
        self, image: array_t, /, *, interval: Tuple[float, float] = None
    ) -> None:
        """"""
        pass

    def UpdateImage(
        self,
        picture: array_t,
        /,
        *,
        interval: Tuple[float, float] = None,
        should_update_limits: bool = False,
    ) -> None:
        """"""
        pass

    def PlotText(self, x, y, s, fontdict=None, **kwargs):
        pass

    def PlotAnnotation(self, text, xy, *args, **kwargs):
        pass


@dtcl.dataclass(init=False, repr=False, eq=False)
class axes_3d_t(axes_t):
    @staticmethod
    def TimeScaling(shape: Sequence[int], length: int, /) -> float:
        """"""
        return 1.0

    @staticmethod
    def MillefeuilleScaling(shape: Sequence[int], length: int, /) -> float:
        """"""
        return 1.0

    def PlotPoints(
        self, xs, ys, zs=0, zdir="z", s=20, c=None, depthshade=True, *args, **kwargs
    ):
        pass

    def PlotIsosurface(
        self,
        volume: array_t,
        iso_value: float,
        /,
        should_be_capped: bool = False,
        keep_every: int = 2,
        **kwargs,
    ) -> None:
        pass

    def PlotImageInZ(
        self,
        image: array_t,
        all_rows: array_t,
        all_cols: array_t,
        n_levels: int,
        height: float,
        z_scaling: float,
        /,
        min_intensity: float = 0.0,
        intensity_range: float = 1.0,
        alpha: float = 0.8,
        **kwargs,
    ) -> None:
        pass

    def PlotText(self, x, y, z, s, zdir=None, **kwargs):
        pass

    def PlotAnnotation(self, text, xyz, *args, **kwargs):
        pass



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

# TODO: this implementation is a temporary solution to plotting division and death events. It should be replaced with
#     e_viewer_2d_t.py in type. This will also be the occasion to extract the axes_track of t_viewer_2d_t, make it
#     something similar to the future e_viewer_2d_t, and turn them both into embeddable axes as is s_viewer_2d_t.

from numbers import Real as number_t
from typing import Callable, Dict, Optional, Sequence, Tuple, Union

import matplotlib.lines as lins
import matplotlib.pyplot as pypl
import numpy as nmpy
from matplotlib.collections import PatchCollection as patches_t
from matplotlib.legend_handler import HandlerTuple as tuple_handler_t
from matplotlib.patches import Rectangle as rectangle_t

from cell_tracking_BC.in_out.text.logger import LOGGER


rectangle_info_h = Tuple[int, float, float, bool, bool]
ValueTransform_h = Optional[Union[float, Callable[[float], float]]]


DIVISION_MARKER = "3"
DEATH_MARKER = "x"
DIVISION_MARKER_SIZE = 60
DEATH_MARKER_SIZE = 30
N_SCORE_LEGEND_SHADES = 8

_RECTANGLE_TYPE_DIV = 0
_RECTANGLE_TYPE_DEA = 1
_RECTANGLE_TYPE_MIXED = 2

_rectangle = rectangle_t((0.0, 0.0), 1.0, 1.0, facecolor="red")
_RED_MATPLOTLIB = _rectangle.get_facecolor()
_rectangle.set_facecolor("blue")
_BLUE_MATPLOTLIB = _rectangle.get_facecolor()
_rectangle.set_facecolor("None")
_NO_COLOR_MATPLOTLIB = _rectangle.get_facecolor()


def ShowDivisionAndDeathEventResponses(
    cell_division_response: Dict[int, Sequence[float]],
    cell_death_response: Dict[int, Sequence[float]],
    cell_division_frame_idc: Dict[int, Optional[Sequence[int]]],
    cell_death_frame_idc: Dict[int, Optional[int]],
    sequence_length: int,
    /,
    *,
    ValueTransform: ValueTransform_h = None,
    zero_is_black: bool = True,
    show_as_barplot: bool = False,
    should_launch_event_loop: bool = True,
) -> None:
    """"""
    if show_as_barplot:
        track_height = 2
    else:
        track_height = 1
    max_label, div_values, dea_values, rectangles_w_details = _ValuesAndRectangles(
        cell_division_response,
        cell_death_response,
        cell_division_frame_idc,
        cell_death_frame_idc,
        show_as_barplot,
    )
    if rectangles_w_details.__len__() == 0:
        LOGGER.info("Empty division or death responses")
        return

    _SetRectanglesColorAndHeight(
        rectangles_w_details,
        div_values,
        dea_values,
        ValueTransform,
        zero_is_black,
        show_as_barplot,
    )
    events = _DivisionAndDeathEvents(rectangles_w_details)

    figure, axes = pypl.subplots()

    rectangles = tuple(_elm[0] for _elm in rectangles_w_details)
    edge_colors = tuple(_rct.get_edgecolor() for _rct in rectangles)
    face_colors = tuple(_rct.get_facecolor() for _rct in rectangles)
    rectangles = patches_t(rectangles, edgecolors=edge_colors, facecolors=face_colors)
    axes.add_collection(rectangles)
    if show_as_barplot:
        color = "k"
    elif zero_is_black:
        color = "w"
    else:
        color = "k"
    for what, size, where in events:
        axes.scatter(*where, c=color, marker=what, s=size)
    _SetAxesProperties(axes, sequence_length, max_label, track_height)
    _AddLegend(axes, zero_is_black)

    figure.tight_layout()
    if should_launch_event_loop:
        pypl.show()


def _ValuesAndRectangles(
    cell_division_response: Dict[int, Sequence[float]],
    cell_death_response: Dict[int, Sequence[float]],
    cell_division_frame_idc: Dict[int, Optional[Sequence[int]]],
    cell_death_frame_idc: Dict[int, Optional[int]],
    show_as_barplot: bool,
    /,
) -> Tuple[
    int,
    Sequence[float],
    Sequence[float],
    Sequence[Tuple[rectangle_t, rectangle_info_h]],
]:
    """"""
    max_label = max(max(cell_division_response.keys()), max(cell_death_response.keys()))
    div_values = []
    dea_values = []
    rectangles = []

    for (
        (label, division_response),
        death_response,
        division_time_points,
        death_frame_idx,
    ) in zip(
        cell_division_response.items(),
        cell_death_response.values(),
        cell_division_frame_idc.values(),
        cell_death_frame_idc.values(),
    ):
        if (division_response is None) and (death_response is None):
            continue

        if division_response is not None:
            div_values.extend(division_response)
        if death_response is not None:
            dea_values.extend(death_response)

        # Whether pattern-based or track-based
        death_occurred = death_frame_idx is not None

        # Could be inside an "if not show_as_barplot", but the IDE might complain about potentially uninitialized object
        rtype = _RECTANGLE_TYPE_MIXED
        if division_response is None:
            if show_as_barplot:
                rtype = _RECTANGLE_TYPE_DEA
            for t_idx, dea_value in enumerate(death_response):
                if dea_value is None:
                    # The trajectory does not start at time point 0
                    continue

                # Note: division_time_points is None if no divisions (used to be (-1,))
                # Note: division is checked despite the track having no division response because the sibling track
                # might have a response (if long enough) higher enough.
                dividing = (division_time_points is not None) and (
                    t_idx in division_time_points
                )
                # Detected by pattern matching (track ending => negative death frame index)
                dying = death_occurred and (t_idx == death_frame_idx)

                rectangle = _EventRectangle(t_idx, label, show_as_barplot)
                details = (rtype, 0.0, dea_value, dividing, dying)
                rectangles.append((rectangle, details))

                if dying:
                    break
        elif death_response is None:
            if show_as_barplot:
                rtype = _RECTANGLE_TYPE_DIV
            for t_idx, div_value in enumerate(division_response):
                if div_value is None:
                    # The trajectory does not start at time point 0
                    continue

                # Note: division_time_points is None if no divisions (used to be (-1,))
                dividing = (division_time_points is not None) and (
                    t_idx in division_time_points
                )

                rectangle = _EventRectangle(t_idx, label, show_as_barplot)
                details = (rtype, div_value, 0.0, dividing, False)
                rectangles.append((rectangle, details))
        else:
            for t_idx, (div_value, dea_value) in enumerate(
                zip(division_response, death_response)
            ):
                if div_value is None:
                    # The trajectory does not start at time point 0
                    if dea_value is not None:
                        raise ValueError(
                            f"{label}: Trajectory has a valid death response at time point {t_idx} "
                            f"whereas the division response is invalid"
                        )
                    continue

                # Note: division_time_points is None if no divisions (used to be (-1,))
                dividing = (division_time_points is not None) and (
                    t_idx in division_time_points
                )
                # Detected by pattern matching (track ending => negative death frame index)
                dying = death_occurred and (t_idx == death_frame_idx)

                rectangle = _EventRectangle(t_idx, label, show_as_barplot)
                if show_as_barplot:
                    details = (_RECTANGLE_TYPE_DEA, 0.0, dea_value, False, dying)
                    rectangles.append((rectangle, details))

                    rectangle = _EventRectangle(t_idx, label, show_as_barplot)
                    details = (_RECTANGLE_TYPE_DIV, div_value, 0.0, dividing, False)
                else:
                    details = (
                        _RECTANGLE_TYPE_MIXED,
                        div_value,
                        dea_value,
                        dividing,
                        dying,
                    )
                rectangles.append((rectangle, details))

                if dying:
                    break

    return max_label, div_values, dea_values, rectangles


def _EventRectangle(
    time_point: int,
    label: int,
    show_as_barplot: bool,
    /,
) -> rectangle_t:
    """"""
    if show_as_barplot:
        corner_y = 2.0 * label - 1.5
    else:
        corner_y = label - 0.5

    return rectangle_t((time_point, corner_y), 1.0, 1.0)


def _SetRectanglesColorAndHeight(
    rectangles: Sequence[Tuple[rectangle_t, rectangle_info_h]],
    div_values: Sequence[float],
    dea_values: Sequence[float],
    ValueTransform: ValueTransform_h,
    zero_is_black: bool,
    show_as_barplot: bool,
    /,
) -> None:
    """
    ValueTransform: if a number, then value <= ValueTransform => value <- 0.0
    """
    if ValueTransform is None:
        ValueTransform = lambda _vle: _vle
    elif isinstance(ValueTransform, number_t):
        threshold = float(ValueTransform)

        def _ValueTransform(_vle: float, /) -> float:
            if _vle <= threshold:
                return 0.0
            else:
                return _vle

        ValueTransform = _ValueTransform

    min_div_value, div_scaling = _MinimumAndScaling(div_values)
    min_dea_value, dea_scaling = _MinimumAndScaling(dea_values)

    for rectangle, (rtype, div_value, dea_value, dividing, dying) in rectangles:
        div = ValueTransform(div_scaling * (div_value - min_div_value))
        dea = ValueTransform(dea_scaling * (dea_value - min_dea_value))

        if show_as_barplot:
            if rtype == _RECTANGLE_TYPE_DEA:
                if zero_is_black:
                    color = (dea, 0.0, 0.0)
                else:
                    color = (1.0, 1.0 - dea, 1.0 - dea)
                rectangle.set_height(1.5 * dea)
                rectangle.set_facecolor(_NO_COLOR_MATPLOTLIB)
            else:
                if zero_is_black:
                    color = (0.0, 0.0, div)
                else:
                    color = (1.0 - div, 1.0 - div, 1.0)
                rectangle.set_height(1.5 * div)
                rectangle.set_facecolor(color + (0.3,))
            rectangle.set_edgecolor(color)
        else:
            if zero_is_black:
                color = (dea, 0.0, div)
            else:
                div_color = (1.0 - div, 1.0 - div, 1.0)
                dea_color = (1.0, 1.0 - dea, 1.0 - dea)
                color = nmpy.minimum(div_color, dea_color)
            rectangle.set_edgecolor(color)
            rectangle.set_facecolor(color)


def _DivisionAndDeathEvents(
    rectangles: Sequence[Tuple[rectangle_t, rectangle_info_h]],
    /,
) -> Sequence[Tuple[str, int, Sequence[float]]]:
    """"""
    output = []

    for rectangle, (*_, dividing, dying) in rectangles:
        if dividing or dying:
            where = tuple(_crd + 0.5 for _crd in rectangle.get_xy())
            if dividing:
                output.append((DIVISION_MARKER, DIVISION_MARKER_SIZE, where))
            if dying:
                output.append((DEATH_MARKER, DEATH_MARKER_SIZE, where))

    return output


def _MinimumAndScaling(values: Sequence[float], /) -> Tuple[float, float]:
    """"""
    if values.__len__() > 0:
        min_value = min(values)
        max_value = max(values)
        scaling = 1.0 / (max_value - min_value)
    else:
        min_value, scaling = 0.0, 1.0

    return min_value, scaling


def _SetAxesProperties(
    axes: pypl.Axes, sequence_length: int, max_label: int, track_height: int, /
) -> None:
    """"""
    axes.set_xlim(left=0, right=sequence_length)
    axes.set_ylim(bottom=0, top=track_height * max_label + 1)

    if track_height == 2:
        offset = -0.5
    else:
        offset = 0.0
    positions = range(1, max_label + 1, max(1, int(round(max_label / 10))))
    axes.yaxis.set_ticks(track_height * nmpy.array(positions) + offset)
    axes.yaxis.set_ticklabels(positions)


def _AddLegend(axes: pypl.Axes, zero_is_black: bool, /) -> None:
    """"""
    shades = nmpy.linspace(0.0, 1.0, num=N_SCORE_LEGEND_SHADES)

    if zero_is_black:
        DivisionColor = lambda _shd: (0.0, 0.0, _shd)
        DeathColor = lambda _shd: (_shd, 0.0, 0.0)
    else:
        DivisionColor = lambda _shd: (1.0 - _shd, 1.0 - _shd, 1.0)
        DeathColor = lambda _shd: (1.0, 1.0 - _shd, 1.0 - _shd)
    score_legends = [
        [
            rectangle_t(
                (0, 0),
                3.0,
                5.0,
                edgecolor=_NO_COLOR_MATPLOTLIB,
                facecolor=_Clr(_shd),
            )
            for _shd in shades
        ]
        for _Clr in (DivisionColor, DeathColor)
    ]

    event_legends = [
        lins.Line2D(
            (),
            (),
            color="k",
            marker=_mrk,
            markersize=_sze,
            linestyle="None",
        )
        for _mrk, _sze in (
            (DIVISION_MARKER, DIVISION_MARKER_SIZE // 5),
            (DEATH_MARKER, DEATH_MARKER_SIZE // 5),
        )
    ]

    axes.legend(
        handles=(*score_legends, *event_legends),
        labels=("Division Score", "Death Score", "Division", "Death (pattern)"),
        loc="right",
        bbox_to_anchor=(1.3, 0.5),
        handler_map={list: tuple_handler_t(ndivide=None, pad=0)},
    )



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from typing import Callable, Optional, Sequence, Tuple, Union

import numpy as nmpy

import cell_tracking_BC.in_out.graphics.generic.d_any as gphc
from cell_tracking_BC.in_out.graphics.type.annotation import annotation_h
from cell_tracking_BC.in_out.graphics.type.axes import axes_2d_t as axes_t
from cell_tracking_BC.in_out.graphics.type.context import context_t
from cell_tracking_BC.in_out.graphics.type.figure import figure_t
from cell_tracking_BC.in_out.graphics.type.signatures import signatures_p
from cell_tracking_BC.in_out.graphics.type.widget import slider_h
from cell_tracking_BC.type.frame import frame_t
from cell_tracking_BC.type.segmentation import segmentation_t
from cell_tracking_BC.type.segmentations import segmentations_t
from cell_tracking_BC.type.sequence import (
    AllChannelsOfSequence,
    AllSegmentationsOfSequence,
    AllStreamsOfSequence,
    all_versions_h,
    sequence_h,
    sequence_t,
)
from cell_tracking_BC.type.tracks import tracks_t


array_t = nmpy.ndarray


@dtcl.dataclass(repr=False, eq=False)
class s_drawer_2d_t:

    figure: figure_t
    axes: axes_t

    all_versions: all_versions_h
    current_version: str = None
    current_time_point: int = -1  # Used only when slider is None
    current_label: int = -1

    slider: slider_h = None

    # Only meaningful for NewForChannels
    cell_contours: Sequence[Sequence[array_t]] = None
    # Using frame labeling for array_t's, or cell_frames below for sequence_t
    with_cell_labels: bool = False
    cell_frames: Sequence[frame_t] = None
    tracks: tracks_t = None
    annotations: Sequence[
        Tuple[int, Union[annotation_h, Sequence[annotation_h]]]
    ] = None

    dbe: context_t = None

    @classmethod
    def NewForChannels(
        cls,
        sequence: Union[Sequence[array_t], sequence_t],
        dbe: context_t,
        /,
        *,
        channel: str = None,
        with_segmentation: bool = False,
        with_cell_labels: bool = False,
        with_track_labels: bool = False,
        in_axes: axes_t = None,
        with_ticks: bool = True,
        with_colorbar: bool = True,
    ) -> s_drawer_2d_t:
        """"""
        with_cell_labels = (
            with_cell_labels and isinstance(sequence, sequence_t) and sequence.has_cells
        )
        if isinstance(sequence, sequence_t) and sequence.has_cells:
            cell_frames = sequence.cell_frames
        else:
            cell_frames = None

        instance = cls._NewForSequence(
            sequence,
            AllChannelsOfSequence,
            dbe,
            version=channel,
            with_segmentation=with_segmentation,
            with_cell_labels=with_cell_labels,
            cell_frames=cell_frames,
            with_track_labels=with_track_labels,
            in_axes=in_axes,
            with_ticks=with_ticks,
        )
        if with_colorbar:
            instance.AddColorbarForImage()

        return instance

    @classmethod
    def NewForSegmentation(
        cls,
        sequence: sequence_h,
        dbe: context_t,
        /,
        *,
        version: str = None,
        with_cell_labels: bool = True,
        with_track_labels: bool = True,
        in_axes: axes_t = None,
        with_ticks: bool = True,
    ) -> s_drawer_2d_t:
        """"""
        if (
            isinstance(sequence, Sequence)
            and isinstance(sequence[0], segmentation_t)
            and not isinstance(sequence, segmentations_t)
        ):
            new_sequence = segmentations_t()
            for segmentation in sequence:
                new_sequence.append(segmentation)
            sequence = new_sequence

        if isinstance(sequence, sequence_t) and sequence.has_cells:
            cell_frames = sequence.cell_frames
        else:
            cell_frames = None

        return cls._NewForSequence(
            sequence,
            AllSegmentationsOfSequence,
            dbe,
            version=version,
            with_cell_labels=with_cell_labels,
            cell_frames=cell_frames,
            with_track_labels=with_track_labels,
            in_axes=in_axes,
            with_ticks=with_ticks,
        )

    @classmethod
    def NewForAllStreams(
        cls,
        sequence: sequence_t,
        dbe: context_t,
        /,
        *,
        version: str = None,
        with_segmentation: bool = False,
        with_cell_labels: bool = False,
        with_track_labels: bool = False,
        in_axes: axes_t = None,
        with_ticks: bool = True,
        with_colorbar: bool = True,
    ) -> s_drawer_2d_t:
        """"""
        with_cell_labels = with_cell_labels and sequence.has_cells
        if sequence.has_cells:
            cell_frames = sequence.cell_frames
        else:
            cell_frames = None

        instance = cls._NewForSequence(
            sequence,
            AllStreamsOfSequence,
            dbe,
            version=version,
            with_segmentation=with_segmentation,
            with_cell_labels=with_cell_labels,
            cell_frames=cell_frames,
            with_track_labels=with_track_labels,
            in_axes=in_axes,
            with_ticks=with_ticks,
        )
        if with_colorbar:
            instance.AddColorbarForImage()

        return instance

    @classmethod
    def _NewForSequence(
        cls,
        sequence: sequence_h,
        AllVersionsOfSequence: Callable[[sequence_h], Tuple[all_versions_h, str]],
        dbe: context_t,
        /,
        *,
        version: str = None,
        with_segmentation: bool = False,
        with_cell_labels: bool = True,
        cell_frames: Sequence[frame_t] = None,
        with_track_labels: bool = True,
        in_axes: axes_t = None,
        with_ticks: bool = True,
    ) -> s_drawer_2d_t:
        """"""
        if in_axes is None:
            figure, axes = dbe.figure_2d_t.NewFigureAndAxes()
        else:
            figure = in_axes.Figure()
            axes = in_axes
        if not with_ticks:
            axes.TurnTicksOff()

        all_versions, current_version = AllVersionsOfSequence(sequence)
        if version is not None:
            current_version = version
        if all_versions.__len__() > 1:
            axes.SetTitle(current_version)

        cell_contours = gphc.CellContours(sequence, with_segmentation)
        tracks = gphc.CellTracks(sequence, with_track_labels)

        cell_annotations = _ShowFirstFrame(
            all_versions,
            current_version,
            cell_contours,
            with_cell_labels,
            cell_frames,
            tracks,
            axes,
            dbe.CellAnnotationStyle,
        )

        instance = cls(
            figure=figure,
            axes=axes,
            all_versions=all_versions,
            current_version=current_version,
            current_time_point=0,
            cell_contours=cell_contours,
            with_cell_labels=with_cell_labels,
            tracks=tracks,
            cell_frames=cell_frames,
            annotations=cell_annotations,
            dbe=dbe,
        )

        return instance

    def AddColorbarForImage(self) -> None:
        """"""
        raise NotImplementedError

    def ShowFrame(
        self,
        /,
        *,
        version: str = None,
        time_point: Union[int, float] = None,
        highlighted: int = -1,
        should_update_limits: bool = False,
        should_draw: bool = True,
        force_update: bool = False,
    ) -> None:
        """
        force_update: If the slider has been updated externally, the time point will not be considered new, and no
        update will be made. Hence, this parameter.
        """
        if version is None:
            version = self.current_version
        if self.slider is None:
            current_time_point = self.current_time_point
        else:
            current_time_point = self.slider.val
        if time_point is None:
            time_point = current_time_point
        else:
            time_point = int(time_point)

        version_is_new = version != self.current_version
        time_point_is_new = (time_point != current_time_point) or force_update

        if version_is_new or time_point_is_new:
            interval, frames = self.all_versions[version]
            frame = frames[time_point]
            self.axes.UpdateImage(
                frame, interval=interval, should_update_limits=should_update_limits
            )
        else:
            frame = None

        if self.annotations is not None:
            if time_point_is_new:
                if self.cell_contours is None:
                    contours = None
                else:
                    contours = self.cell_contours[time_point]
                if self.cell_frames is None:
                    cell_frame = None
                else:
                    cell_frame = self.cell_frames[time_point]

                self.annotations = self.axes.PlotCellsDetails(
                    frame,
                    contours,
                    self.with_cell_labels,
                    cell_frame,
                    self.tracks,
                    self.dbe.CellAnnotationStyle,
                    highlighted=highlighted,
                )
                self.current_label = highlighted
            elif highlighted > 0:
                self.HighlightAnnotation(highlighted, should_draw=False)

        if version_is_new:
            self.axes.SetTitle(version)
            self.current_version = version

        if time_point_is_new:
            if self.slider is None:
                self.current_time_point = time_point
            else:
                self.dbe.UpdateSlider(self.slider, time_point)

        if should_draw:
            self.figure.Update()

    def HighlightAnnotation(self, label: int, /, *, should_draw: bool = True) -> None:
        """
        If label is <= 0 or > max cell label in current frame, then un-highlights all annotations
        """
        raise NotImplementedError


def _ShowFirstFrame(
    all_versions: all_versions_h,
    current_version: str,
    cell_contours: Optional[Sequence[Sequence[array_t]]],
    with_cell_labels: bool,
    cell_frames: Optional[Sequence[frame_t]],
    tracks: Optional[tracks_t],
    axes: axes_t,
    CellAnnotationStyle: signatures_p.cell_annotation_style_h,
    /,
) -> Optional[Sequence[Tuple[int, annotation_h]]]:
    """"""
    interval, version = all_versions[current_version]
    first_frame = version[0]

    axes.PlotImage(first_frame, interval=interval)

    if (cell_contours is not None) or with_cell_labels or (tracks is not None):
        if cell_contours is None:
            contours = None
        else:
            contours = cell_contours[0]
        if cell_frames is None:
            cell_frame = None
        else:
            cell_frame = cell_frames[0]
        cell_annotations = axes.PlotCellsDetails(
            first_frame,
            contours,
            with_cell_labels,
            cell_frame,
            tracks,
            CellAnnotationStyle,
        )
    else:
        cell_annotations = None

    # Once the first frame has been plot, disable axes autoscale to try to speed future plots up
    axes.Freeze()

    return cell_annotations



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Sequence, Union


ISSUE_SEPARATOR = ": "


def FactorizedIssuesAsStr(
    issue: Union[str, Sequence[str]], /, *, max_n_contexts: int = 1000
) -> str:
    """"""
    if isinstance(issue, str):
        return issue

    issues = issue
    issues_per_content = {}
    for issue in issues:
        if ISSUE_SEPARATOR in issue:
            context, content = issue.split(ISSUE_SEPARATOR, 1)
        else:
            context, content = "", issue
        if content in issues_per_content:
            issues_per_content[content].append(context)
        else:
            issues_per_content[content] = [context]

    factorized = []
    first_half = max_n_contexts // 2
    last_half = max_n_contexts - first_half - 1
    for content, contexts in issues_per_content.items():
        contexts = tuple(_ctx for _ctx in contexts if _ctx.__len__() > 0)
        if (n_contexts := contexts.__len__()) > 1:
            if n_contexts > max_n_contexts:
                contexts = contexts[:first_half] + ("...",) + contexts[(-last_half):]
            contexts_as_str = " / ".join(contexts)
            factorized.append(f"{content}{ISSUE_SEPARATOR}{contexts_as_str}")
        elif n_contexts > 0:
            factorized.append(f"{contexts[0]}{ISSUE_SEPARATOR}{content}")
        else:
            factorized.append(content)

    return " | ".join(factorized)



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
import itertools as ittl
from operator import attrgetter as GetAttribute
from operator import itemgetter as ItemAt


from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

import networkx as grph
from numpy import ndarray as array_t

from cell_tracking_BC.standard.number import INFINITY_MINUS
from cell_tracking_BC.standard.uid import ShortID
from cell_tracking_BC.type.cell import cell_t
from cell_tracking_BC.type.track import (
    TIME_POINT,
    any_track_h,
    forking_track_t,
    single_track_t,
    structured_track_h,
    unstructured_track_t,
)


invalid_tracks_h = Dict[
    Type[any_track_h], List[Tuple[any_track_h, Union[str, Sequence[str]]]]
]
per_single_track_cells = Dict[int, Sequence[Union[cell_t, Tuple[cell_t, int]]]]
TrackIssues_h = Callable[
    [structured_track_h, dict], Optional[Union[str, Sequence[str]]]
]
DividingCells_h = Callable[
    [forking_track_t, bool], Sequence[Union[cell_t, Tuple[cell_t, int]]]
]


def _Invalids() -> invalid_tracks_h:
    """"""
    return {
        unstructured_track_t: [],
        single_track_t: [],
        forking_track_t: [],
    }


@dtcl.dataclass(repr=False, eq=False)
class tracks_t(List[structured_track_h]):

    invalids: invalid_tracks_h = dtcl.field(init=False, default_factory=_Invalids)

    @classmethod
    def NewFromUnstructuredTracks(cls, tracks: unstructured_tracks_t, /) -> tracks_t:
        """"""
        instance = cls()

        next_single_track_label = 1
        for unstructured in tracks.track_iterator:
            issues = unstructured.Issues()
            if issues is None:
                (
                    forking_track,
                    next_single_track_label,
                ) = forking_track_t.NewFromUnstructuredTrack(
                    unstructured, next_single_track_label
                )

                if forking_track.n_leaves > 1:
                    instance.append(forking_track)
                else:
                    single_track = forking_track.AsSingleTrack()
                    instance.append(single_track)
            else:
                # unstructured.graph["issues"] = issues  # Old, per-track issue storage
                instance.invalids[unstructured_track_t].append((unstructured, issues))

        return instance

    def RootCells(
        self, /, *, with_time_point: bool = False
    ) -> Sequence[Union[cell_t, Tuple[cell_t, int]]]:
        """"""
        if with_time_point:
            output = map(GetAttribute("root", "root_time_point"), self)
            # output = ((_tck.root, _tck.root_time_point) for _tck in self)
        else:
            output = map(GetAttribute("root"), self)
            # output = (_tck.root for _tck in self)

        return tuple(output)

    def DividingCells(
        self,
        /,
        *,
        with_time_point: bool = False,
        per_single_track: bool = False,
    ) -> Union[Sequence[Union[cell_t, Tuple[cell_t, int]]], per_single_track_cells,]:
        """"""
        output = []

        for track in self:
            if isinstance(track, forking_track_t):
                dividing = track.DividingCells(with_time_point=with_time_point)
                output.extend(dividing)

        if per_single_track:
            per_track = {}
            for track in self.single_tracks_iterator:
                if with_time_point:
                    dividing = filter(lambda _elm: _elm[0] in track, output)
                    dividing = sorted(dividing, key=ItemAt(1))
                else:
                    dividing = tuple(filter(lambda _elm: _elm in track, output))
                per_track[track.label] = dividing

            output = per_track
        elif with_time_point:
            output.sort(key=ItemAt(1))

        return output

    def FilteredDividingCells(
        self,
        division_responses: Dict[int, Optional[Union[array_t, Sequence[float]]]],
        lower_bound: float,
        /,
        *,
        with_time_point: bool = False,
        should_return_discarded: bool = False,
    ) -> Union[
        per_single_track_cells, Tuple[per_single_track_cells, per_single_track_cells]
    ]:
        """
        As a per-single-track dictionary
        """
        output_1 = {}
        output_2 = {}

        unfiltered = self.DividingCells(with_time_point=True, per_single_track=True)
        for label, divisions in unfiltered.items():
            retained = []
            discarded = []
            for cell, time_point in divisions:
                sibling_labels = self.TrackLabelsContainingCell(cell)
                highest_response = max(
                    division_responses[_lbl][time_point]
                    if (division_responses[_lbl] is not None)
                    and (division_responses[_lbl][time_point] is not None)
                    else INFINITY_MINUS
                    for _lbl in sibling_labels
                )
                if highest_response >= lower_bound:
                    if with_time_point:
                        retained.append((cell, time_point))
                    else:
                        retained.append(cell)
                elif should_return_discarded:
                    if with_time_point:
                        discarded.append((cell, time_point))
                    else:
                        discarded.append(cell)

            output_1[label] = tuple(retained)
            output_2[label] = tuple(discarded)

        if should_return_discarded:
            return output_1, output_2

        return output_1

    def LeafCells(
        self, /, *, with_time_point: bool = False
    ) -> Sequence[Union[cell_t, Tuple[cell_t, int]]]:
        """"""
        leaves = []
        time_points = []
        for track in self:
            leaves.extend(track.leaves)
            time_points.extend(track.leaves_time_points)

        if with_time_point:
            return tuple(zip(leaves, time_points))

        return tuple(leaves)

    def TrackWithRoot(
        self, root: cell_t, /, *, tolerant_mode: bool = False
    ) -> Optional[structured_track_h]:
        """"""
        for track in self:
            if root is track.root:
                return track

        if tolerant_mode:
            return None

        raise ValueError(f"{root}: Not a root cell")

    def TrackWithLeaf(
        self, leaf: cell_t, /, *, tolerant_mode: bool = False
    ) -> Optional[single_track_t]:
        """
        TrackWithLeaf: Implicitly, it is SingleTrackWithLeaf
        """
        for track in self:
            for cell in track.leaves:
                if leaf is cell:
                    return track.TrackWithLeaf(cell, check=False)

        if tolerant_mode:
            return None

        raise ValueError(f"{leaf}: Not a leaf cell")

    @property
    def single_tracks_iterator(self) -> Iterator[single_track_t]:
        """"""
        for track in self:
            for single_track in track.single_tracks_iterator:
                yield single_track

    def TrackLabelsContainingCell(
        self, cell: cell_t, /, *, tolerant_mode: bool = False
    ) -> Optional[Sequence[int]]:
        """"""
        for track in self:
            if cell in track:
                return track.TrackLabelsContainingCell(cell, check=False)

        if tolerant_mode:
            return None

        raise ValueError(f"{cell}: Not a tracked cell")

    def TrackLabelWithLeaf(
        self, leaf: cell_t, /, *, tolerant_mode: bool = False
    ) -> Optional[int]:
        """
        TrackLabelWithLeaf: Implicitly, it is SingleTrackLabelWithLeaf
        """
        track = self.TrackWithLeaf(leaf, tolerant_mode=True)
        if track is not None:
            return track.label

        if tolerant_mode:
            return None

        raise ValueError(f"{leaf}: Not a leaf cell")

    def Prune(
        self,
        ShouldStartPruning: Callable[[int, cell_t, dict], bool],
        /,
        *,
        ShouldStopPruning: Callable[[int, cell_t, dict], bool] = None,
        start_arguments: Dict[str, Any] = None,
        stop_arguments: Dict[str, Any] = None,
    ) -> None:
        """"""
        t_idx = 0
        while t_idx < self.__len__():
            track = self[t_idx]
            track.Prune(
                ShouldStartPruning,
                ShouldStopPruning=ShouldStopPruning,
                start_arguments=start_arguments,
                stop_arguments=stop_arguments,
                called_from_educated_code=True,
            )
            if track.IsFullyPruned():
                self.invalids[track.__class__].append((track, "Fully pruned"))
                del self[t_idx]
            else:
                t_idx += 1

    def FilterOut(
        self,
        TrackIssues: TrackIssues_h,
        /,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        TrackIssues: Arguments are: track and (optional) keyword arguments; Returned value can be None, an str, or a
            sequence of str.
        kwargs: Passed to TrackIsInvalid as keyword arguments
        """
        t_idx = 0
        while t_idx < self.__len__():
            track = self[t_idx]
            issues = TrackIssues(track, **kwargs)
            if issues is None:
                t_idx += 1
            else:
                self.invalids[track.__class__].append((track, issues))
                del self[t_idx]

    def Print(self) -> None:
        """"""
        for track in self:
            print(track)

    def __str__(self) -> str:
        """"""
        return (
            f"{self.__class__.__name__.upper()}.{ShortID(id(self))}: {self.__len__()=}"
        )


class unstructured_tracks_t(grph.DiGraph):
    def AddTrackSegment(
        self,
        src_cell: cell_t,
        tgt_cell: cell_t,
        src_time_point: int,
        affinity: float,
        /,
    ) -> None:
        """"""
        time_point = {TIME_POINT: src_time_point}
        time_point_p_1 = {TIME_POINT: src_time_point + 1}
        self.add_node(src_cell, **time_point)
        self.add_node(tgt_cell, **time_point_p_1)
        self.add_edge(src_cell, tgt_cell, affinity=affinity)

    @property
    def track_iterator(self) -> Iterator[unstructured_track_t]:
        """"""
        for cells in grph.weakly_connected_components(self):
            track_view = self.subgraph(cells)
            # Copy or re-instantiation is necessary since the subgraph is a view
            yield unstructured_track_t(track_view)


def UniqueCellsFromPerSingleTrackOnes(
    per_single_track: per_single_track_cells,
    /,
    *,
    should_remove_time_point: bool = False,
) -> Sequence[Union[cell_t, Tuple[cell_t, int]]]:
    """"""
    cells = ittl.chain.from_iterable(per_single_track.values())
    if should_remove_time_point:
        cells = map(ItemAt(0), cells)

    return tuple(set(cells))



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from typing import Sequence, Tuple, Union

import matplotlib.pyplot as pypl
import numpy as nmpy
import scipy.interpolate as ntrp
import skimage.measure as msre
from matplotlib.text import Annotation as annotation_t
from mpl_toolkits.mplot3d import Axes3D as pypl_axes_t

from cell_tracking_BC.in_out.graphics.dbe.matplotlib.d_any import axes_t as axes_anyd_t
from cell_tracking_BC.in_out.graphics.dbe.matplotlib.d_any import (
    figure_t as figure_anyd_t,
)
from cell_tracking_BC.in_out.graphics.type.annotation import cell_annotation_h
from cell_tracking_BC.in_out.graphics.type.axes import axes_3d_t as base_axes_3d_t


array_t = nmpy.ndarray


_MC_STEP_SIZE = 5  # MC=marching cubes
_MIN_MC_TIME_STEPS = 3


@dtcl.dataclass(init=False, repr=False, eq=False)
class axes_t(pypl_axes_t, axes_anyd_t, base_axes_3d_t):

    name = "ctBC_three_d_axes"
    PlotPoints = pypl_axes_t.scatter
    PlotLines = pypl_axes_t.plot
    PlotText = pypl_axes_t.text

    def SetTimeAxisProperties(self, latest: int, /) -> None:
        """"""
        self.set_zlim(0, latest)

        n_ticks = min(20, latest + 1)
        self.set_zticks(nmpy.linspace(0.0, latest, num=n_ticks))
        self.set_zticklabels(
            str(round(latest * _idx / (n_ticks - 1), 1)) for _idx in range(n_ticks)
        )

    def PlotCellAnnotation(
        self, position: Tuple[float, float, float], text: cell_annotation_h, /, **kwargs
    ) -> Union[annotation_t, Sequence[annotation_t]]:
        """"""
        # TODO: multi-piece text is currently dealt with as single-piece text
        if not isinstance(text, str):
            text = "".join(_elm[0] for _elm in text)

        # annotate: only for 2-D; use text instead in 3-D
        output = self.text(*position, text, **kwargs)

        return output

    def PlotImageInZ(
        self,
        image: array_t,
        all_rows: array_t,
        all_cols: array_t,
        n_levels: int,
        height: float,
        z_scaling: float,
        /,
        min_intensity: float = 0.0,
        intensity_range: float = 1.0,
        alpha: float = 0.8,
        **kwargs,
    ) -> None:
        """"""
        self.contourf(
            all_rows,
            all_cols,
            image,
            levels=n_levels,
            offset=height,
            alpha=alpha,
            **kwargs,
        )

    def PlotIsosurface(
        self,
        volume: array_t,
        iso_value: float,
        /,
        should_be_capped: bool = False,
        keep_every: int = 2,
        **kwargs,
    ) -> None:
        """"""
        n_frames_all = volume.shape[0]
        n_frames_kept = int((n_frames_all - 1) / keep_every) + 1
        frame_shape = volume.shape[1:]

        original_extents = (
            range(n_frames_all),
            range(frame_shape[0]),
            range(frame_shape[1]),
        )
        interpolated_extents = (
            nmpy.linspace(
                0,
                n_frames_all - ((n_frames_all - 1) % keep_every) - 1,
                num=n_frames_kept,
            ),
            *original_extents[1:],
        )
        all_times, all_rows, all_cols = nmpy.meshgrid(
            *interpolated_extents, indexing="ij"
        )
        interpolated_sites = nmpy.vstack(
            (all_times.flat, all_rows.flat, all_cols.flat)
        ).T
        interpolated = ntrp.interpn(original_extents, volume, interpolated_sites)

        reshaped = nmpy.reshape(interpolated, (n_frames_kept, *frame_shape))
        reorganized = nmpy.moveaxis(reshaped, (0, 1, 2), (2, 0, 1))
        flipped = nmpy.flip(reorganized, axis=2)
        if should_be_capped:
            cap = nmpy.zeros_like(flipped[..., 0])[..., None]
            flipped = nmpy.dstack((cap, flipped, cap))

        if n_frames_kept / _MC_STEP_SIZE < _MIN_MC_TIME_STEPS:
            step_size = max(1, int(round(n_frames_kept / _MIN_MC_TIME_STEPS)))
        else:
            step_size = _MC_STEP_SIZE
        try:
            vertices, faces, *_ = msre.marching_cubes(
                flipped, iso_value, step_size=step_size
            )
        except RuntimeError as exception:
            print(
                f"{self.__class__.PlotIsosurface.__name__}: Error in {iso_value}-isosurface extraction "
                f"in volume with min {nmpy.amin(flipped)} and max {nmpy.amax(flipped)} "
                f"with step size {step_size}\n{exception}"
            )
            return

        vertices[:, 2] *= keep_every
        self.plot_trisurf(
            vertices[:, 0],
            vertices[:, 1],
            faces,
            nmpy.amax(vertices[:, 2]) - vertices[:, 2],
            **kwargs,
        )


@dtcl.dataclass(init=False, repr=False, eq=False)
class figure_t(figure_anyd_t):
    @classmethod
    def NewFigureAndAxes(
        cls, /, *, n_rows: int = 1, n_cols: int = 1, title: str = None, offline_version: bool = False,
    ) -> Tuple[figure_t, Union[axes_t, Sequence[axes_t], Sequence[Sequence[axes_t]]],]:
        """"""
        figure = pypl.figure(FigureClass=cls)
        axes = axes_t(figure, auto_add_to_figure=False)
        figure.add_axes(axes)

        if title is not None:
            axes.set_title(title)
        axes.set_xlabel("row positions")
        axes.set_ylabel("column positions")
        axes.set_zlabel("time points")

        return figure, axes



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from string import ascii_uppercase as ASCII_UPPERCASE
from typing import Tuple


MIN_ASCII = 33
MAX_ASCII = 126


def ShortID(python_id: int, /) -> str:
    """"""
    output = []

    digits = tuple(int(_dgt) for _dgt in tuple(str(python_id)))
    n_digits = digits.__len__()

    d_idx = 0
    ascii_value = 0
    while d_idx < n_digits:
        new_ascii = 10 * ascii_value + digits[d_idx]
        if new_ascii > MAX_ASCII:
            if ascii_value >= MIN_ASCII:
                output.append(chr(ascii_value))
            else:
                output.append(str(ascii_value))
            while (d_idx < n_digits) and (digits[d_idx] == 0):
                output.append("0")
                d_idx += 1
            if d_idx < n_digits:
                ascii_value = digits[d_idx]
            else:
                ascii_value = None
        else:
            ascii_value = new_ascii
        d_idx += 1

    if ascii_value is not None:
        if ascii_value >= MIN_ASCII:
            output.append(chr(ascii_value))
        else:
            output.append(str(ascii_value))

    return "".join(output)


def AlphaColumnFromIndex(index: int) -> str:
    """
    index: 0-based
    """
    return AlphaColumnFromLabel(index + 1)


def AlphaColumnFromLabel(label: int) -> str:
    """
    label: 1-based

    Adapted from: https://stackoverflow.com/questions/48983939/convert-a-number-to-excel-s-base-26
    Answer: poke, Feb 26 '18 at 9:00 (https://stackoverflow.com/a/48984697)
    """

    def _DivmodExcel(number: int) -> Tuple[int, int]:
        quotient, remainder = divmod(number, 26)
        if remainder > 0:
            return quotient, remainder
        else:
            return quotient - 1, remainder + 26

    output = []

    while label > 0:
        label, digit = _DivmodExcel(label)
        output.append(ASCII_UPPERCASE[digit - 1])

    return "".join(reversed(output))



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Any, Dict

try:
    # noinspection PyPackageRequirements
    import vedo.colors as clrs
except ModuleNotFoundError:
    clrs = None

import cell_tracking_BC.in_out.graphics.dbe.vedo.missing as vmsg
from cell_tracking_BC.in_out.graphics.type.color import ColorAndAlpha


FONT_SIZE_DEFAULT = 14
FONT_SIZE_CONVERSION = {
    "xx-small": 8,
    "x-small": 10,
    "small": 12,
    "medium": 14,
    "large": 16,
    "x-large": 18,
    "xx-large": 20,
}

CELL_ANNOTATION_COLOR_DEFAULT = "red"
CELL_ANNOTATION_COLOR_HIGHLIGHT = "magenta"
CELL_ANNOTATION_STYLE_DEFAULT = {
    "c": CELL_ANNOTATION_COLOR_DEFAULT,
    "s": 8,
    "depth": 1,
    "justify": "centered",
    "alpha": 0.95,
}


def CellAnnotationStyle(highlighted: bool, _: bool, /) -> Dict[str, Any]:
    """"""
    output = CELL_ANNOTATION_STYLE_DEFAULT.copy()

    if highlighted:
        output["c"] = CELL_ANNOTATION_COLOR_HIGHLIGHT
    else:
        output["c"] = CELL_ANNOTATION_COLOR_DEFAULT

    return output


if clrs is None:
    ConvertedTextStyle = vmsg.RaiseMissingVedoException
else:
    def ConvertedTextStyle(style: Dict[str, Any], /) -> Dict[str, Any]:
        """"""
        output = style.copy()

        if "color" in output:
            color, _ = ColorAndAlpha(output["color"], clrs.getColor)
            output["c"] = color
            del output["color"]

        if "fontsize" in output:
            key_or_size = output["fontsize"]
            size = FONT_SIZE_CONVERSION.get(key_or_size, key_or_size)
            output["s"] = size
            del output["fontsize"]

        if "bbox" in output:
            del output["bbox"]

        return output



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dcls
import functools as fctl
from itertools import starmap, zip_longest
from multiprocessing import Pool as pool_t
from operator import attrgetter as GetAttribute
from pathlib import Path as path_t
from typing import Any, Callable, Dict, Iterator, Optional, Sequence, Tuple, Union

import numpy as nmpy
import scipy.ndimage as imge

import cell_tracking_BC.in_out.text.progress as prgs
import cell_tracking_BC.standard.issue as isse
from cell_tracking_BC.standard.number import MAX_INT
from cell_tracking_BC.standard.uid import ShortID
from cell_tracking_BC.type.cell import cell_t
from cell_tracking_BC.type.cytoplasm import cytoplasm_t
from cell_tracking_BC.type.frame import frame_t, transform_h
from cell_tracking_BC.type.nucleus import nucleus_t
from cell_tracking_BC.type.segmentation import compartment_t, segmentation_t
from cell_tracking_BC.type.segmentations import segmentations_t
from cell_tracking_BC.type.track import (
    forking_track_t,
    single_track_t,
    unstructured_track_t,
)
from cell_tracking_BC.type.tracks import tracks_t


array_t = nmpy.ndarray
#
all_versions_h = Dict[
    str, Tuple[Tuple[int, int], Union[Sequence[array_t], Sequence[frame_t]]]
]
channel_computation_h = Callable[[Dict[str, array_t], Dict[str, Any]], array_t]
feature_filtering_h = Callable[..., Sequence[float]]
morphological_feature_computation_h = Callable[
    [cell_t, Dict[str, Any]], Union[Any, Sequence[Any]]
]
radiometric_feature_computation_h = Callable[
    [cell_t, Union[array_t, Sequence[array_t]], Dict[str, Any]],
    Union[Any, Sequence[Any]],
]


@dcls.dataclass(repr=False, eq=False)
class sequence_t:

    path: Optional[path_t] = None
    shape: Tuple[int, int] = None
    original_length: int = None
    length: int = None
    base_channels: Sequence[str] = None
    frames_of_channel: Dict[str, Sequence[Union[array_t, frame_t]]] = None
    cell_channel: str = None  # Name of channel whose frames store the segmented cells
    segmentations: segmentations_t = dcls.field(init=False, default=None)
    tracks: tracks_t = dcls.field(init=False, default=None)

    @classmethod
    def NewFromFrames(
        cls,
        frames: array_t,
        in_channel_names: Sequence[Optional[str]],
        path: path_t,
        /,
        *,
        first_frame: int = 0,
        last_frame: int = MAX_INT,
    ) -> sequence_t:
        """
        in_channel_names: names equal to None or "___" or "---" indicate channels that should be discarded
        """
        # TODO: make this function accept various input shapes thanks to an additional arrangement parameter of the
        #     form THRC, T=time, H=channel, RC= row column. This requires that SequenceFromPath deals with TH combined
        #     dimension.
        if (n_dims := frames.ndim) not in (3, 4):
            raise ValueError(
                f"{n_dims}: Invalid number of dimensions of sequence with shape {frames.shape}; "
                f"Expected=3 or 4=(TIME POINTS*CHANNELS)xROWSxCOLUMNS or "
                f"TIME POINTSxCHANNELSxROWSxCOLUMNS"
            )

        n_in_channels = in_channel_names.__len__()

        frames_of_channel = {}
        for name in in_channel_names:
            if (name is not None) and (name != "___") and (name != "---"):
                frames_of_channel[name] = []
        base_channel_names = tuple(frames_of_channel.keys())

        if n_dims == 3:
            c_idx = n_in_channels - 1
            time_point = -1
            for raw_frame in frames:
                c_idx += 1
                if c_idx == n_in_channels:
                    c_idx = 0
                    time_point += 1

                if time_point < first_frame:
                    continue
                elif time_point > last_frame:
                    break

                name = in_channel_names[c_idx]
                if name in base_channel_names:
                    frame = frame_t(raw_frame)
                    frames_of_channel[name].append(frame)
        else:
            for time_point, raw_frame in enumerate(frames):
                if time_point < first_frame:
                    continue
                elif time_point > last_frame:
                    break

                for c_idx, channel in enumerate(raw_frame):
                    name = in_channel_names[c_idx]
                    if name in base_channel_names:
                        frame = frame_t(channel)
                        frames_of_channel[name].append(frame)

        frames_of_base_channel = frames_of_channel[base_channel_names[0]]
        shape = frames_of_base_channel[0].shape
        length = frames_of_base_channel.__len__()
        instance = cls(
            path=path,
            shape=shape,
            original_length=frames.__len__(),
            length=length,
            base_channels=base_channel_names,
            frames_of_channel=frames_of_channel,
        )

        return instance

    def __len__(self) -> int:
        """"""
        return self.length

    @property
    def channels(self) -> Sequence[str]:
        """
        Names of channels read from file (base channels) and computed channels
        """
        return tuple(self.frames_of_channel.keys())

    def Frames(
        self,
        /,
        *,
        channel: Union[str, Sequence[str]] = None,
    ) -> Union[
        Sequence[Union[array_t, frame_t]],
        Iterator[Sequence[Union[array_t, frame_t]]],
    ]:
        """
        channel: None=all (!) base channels; Otherwise, only (a) base channel name(s) can be passed
        as_iterator: Always considered True if channel is None or a sequence of channel names
        """
        if isinstance(channel, str):
            return self.frames_of_channel[channel]
        else:
            return self._FramesForMultipleChannels(channel)

    def _FramesForMultipleChannels(
        self, channels: Optional[Sequence[str]], /
    ) -> Iterator[Sequence[Union[array_t, frame_t]]]:
        """
        /!\ If "channels" contains both base and non-base channels, then the returned tuples will contain both array_t
        and frame_t elements (but frame_t is a subclass of array_t, so...).
        """
        if channels is None:
            channels = self.base_channels

        for f_idx in range(self.length):
            frames = (self.frames_of_channel[_chl][f_idx] for _chl in channels)

            yield tuple(frames)

    def ChannelExtrema(self, channel: str, /) -> Tuple[float, float]:
        """"""
        min_intensity = nmpy.Inf
        max_intensity = -nmpy.Inf

        for frame in self.frames_of_channel[channel]:
            min_intensity = min(min_intensity, nmpy.amin(frame))
            max_intensity = max(max_intensity, nmpy.amax(frame))

        return min_intensity.item(), max_intensity.item()

    @property
    def has_cells(self) -> bool:
        """"""
        return self.cell_channel is not None

    def NCells(
        self, /, *, in_frame: Union[int, Sequence[int]] = None
    ) -> Union[int, Sequence[int]]:
        """
        in_frame: None=>total over the sequence
        """
        if in_frame is None:
            return sum(_cll.__len__() for _cll in self.cells_iterator)

        just_one = isinstance(in_frame, int)
        if self.has_cells:
            if just_one:
                in_frame = (in_frame,)

            output = in_frame.__len__() * [0]

            for f_idx, cells in enumerate(self.cells_iterator):
                if f_idx in in_frame:
                    output[in_frame.index(f_idx)] = cells.__len__()

            if just_one:
                output = output[0]
        elif just_one:
            output = 0
        else:
            output = in_frame.__len__() * (0,)

        return output

    @property
    def cell_frames(self) -> Sequence[frame_t]:
        """"""
        return self.frames_of_channel[self.cell_channel]

    @property
    def cells_iterator(self) -> Iterator[Sequence[cell_t]]:
        """"""
        for frame in self.cell_frames:
            yield frame.cells

    @property
    def cytoplasms_iterator(self) -> Iterator[Sequence[Optional[cytoplasm_t]]]:
        """"""
        for frame in self.cell_frames:
            yield tuple(map(GetAttribute("cytoplasm"), frame.cells))
            # output = []
            # for cell in frame.cells:
            #     output.append(cell.cytoplasm)
            # yield output

    @property
    def nuclei_iterator(self) -> Iterator[Sequence[Sequence[nucleus_t]]]:
        """"""
        for frame in self.cell_frames:
            yield tuple(map(GetAttribute("nuclei"), frame.cells))
            # output = []
            # for cell in frame.cells:
            #     output.append(cell.nuclei)
            # yield output

    def ApplyTransform(
        self,
        Transform: transform_h,
        /,
        *,
        channel: Union[str, Sequence[str]] = None,
        **kwargs,
    ) -> None:
        """
        channel: None=all (!)
        """
        if channel is None:
            channels = self.base_channels
        elif isinstance(channel, str):
            channels = (channel,)
        else:
            channels = channel

        for channel in channels:
            targets = self.frames_of_channel[channel]
            references_sets = self.Frames(channel=self.channels)
            for target, references in zip(targets, references_sets):
                refs_as_dict = dict(zip(self.channels, references))
                # refs_as_dict = {
                #     _nme: _frm for _nme, _frm in zip(self.channels, references)
                # }
                target.ApplyTransform(Transform, channels=refs_as_dict, **kwargs)

    def AddComputedChannel(
        self, name: str, ChannelComputation: channel_computation_h, /, **kwargs
    ) -> None:
        """"""
        computed = []
        for frames in self.Frames(channel=self.channels):
            frames_as_dict = dict(zip(self.channels, frames))
            # frames_as_dict = {_nme: _frm for _nme, _frm in zip(self.channels, frames)}
            computed.append(ChannelComputation(frames_as_dict, **kwargs))

        self.frames_of_channel[name] = computed

    def AddCellsFromSegmentations(
        self,
        channel: str,
        segmentations: segmentations_t,
    ) -> None:
        """
        Segmentation are supposed to be binary (as opposed to already labeled)
        """
        self.cell_channel = channel
        self.segmentations = segmentations

        for frame, segmentation in zip(self.cell_frames, segmentations):
            frame.AddCellsFromSegmentation(segmentation)

    def AddCellFeature(
        self,
        name: Union[str, Sequence[str]],
        Feature: Union[
            morphological_feature_computation_h, radiometric_feature_computation_h
        ],
        /,
        channel: Union[str, Sequence[str]] = None,
        should_run_in_parallel: bool = True,
        should_run_silently: bool = False,
        **kwargs,
    ) -> None:
        """
        name: If an str, then the value returned by Feature will be considered as a whole, whether it is actually a
        single value or a value container. If a sequence of str's, then the object returned by Feature will be iterated
        over, each element being matched with the corresponding name in "name".
        channel: if None, then morphological feature, else radiometric feature.
        """
        if isinstance(name, str):
            description = f'Feature "{name}"'
        else:
            description = f'Feature "{name[0]}, ..., {name[-1]}"'
        ParallelFeature = fctl.partial(Feature, **kwargs)

        with prgs.ProgressDesign(should_be_silent=should_run_silently) as progress:
            if channel is None:
                iterator = self.cells_iterator
            else:
                iterator = zip(self.cells_iterator, self.frames_of_channel[channel])
            progress_context = prgs.progress_context_t(
                progress,
                iterator,
                total=self.length,
                description=description,
            )

            if should_run_in_parallel:
                pool = pool_t()
                MapFunctionOnList = pool.map
                StarMapFunctionOnList = pool.starmap
            else:
                pool = None
                MapFunctionOnList = map
                StarMapFunctionOnList = starmap

            if channel is None:
                if isinstance(name, str):
                    for cells in progress_context.elements:
                        features = MapFunctionOnList(ParallelFeature, cells)
                        for cell, feature in zip(cells, features):
                            cell.AddFeature(name, feature)
                else:
                    names = name
                    for cells in progress_context.elements:
                        multi_features = MapFunctionOnList(ParallelFeature, cells)
                        for cell, features in zip(cells, multi_features):
                            for name, feature in zip(names, features):
                                cell.AddFeature(name, feature)
            else:
                if isinstance(name, str):
                    for cells, frame in progress_context.elements:
                        features = StarMapFunctionOnList(
                            ParallelFeature,
                            zip_longest(cells, (frame,), fillvalue=frame),
                        )
                        for cell, feature in zip(cells, features):
                            cell.AddFeature(name, feature)
                else:
                    names = name
                    for cells, frame in progress_context.elements:
                        multi_features = StarMapFunctionOnList(
                            ParallelFeature,
                            zip_longest(cells, (frame,), fillvalue=frame),
                        )
                        for cell, features in zip(cells, multi_features):
                            for name, feature in zip(names, features):
                                cell.AddFeature(name, feature)

            if should_run_in_parallel:
                pool.close()
                pool.terminate()

    def FilteredCellFeature(
        self,
        feature: Union[str, Sequence[str]],
        Filter: feature_filtering_h,
        /,
        *args,
        **kwargs,
    ) -> Dict[int, Sequence[Any]]:
        """"""
        output = {}

        if isinstance(feature, str):
            features = (feature,)
        else:
            features = feature
        if any(_ftr not in self.available_cell_features for _ftr in features):
            raise ValueError(f"{feature}: Invalid feature(s)")

        for track in self.tracks.single_tracks_iterator:
            all_series = [
                self.FeatureEvolutionAlongTrack(track, _ftr) for _ftr in features
            ]
            filtered = Filter(*all_series, *args, **kwargs)
            if filtered is None:
                output[track.label] = None
            else:
                pre_padding = track.root_time_point * (None,)
                filtered = pre_padding + tuple(filtered)
                output[track.label] = filtered

        return output

    @property
    def available_cell_features(self) -> Sequence[str]:
        """"""
        one_cell = self.cell_frames[0].cells[0]

        return one_cell.available_features

    def AddTracks(
        self,
        tracks: tracks_t,
        /,
    ) -> None:
        """"""
        self.tracks = tracks

    @staticmethod
    def FeatureEvolutionAlongTrack(
        track: single_track_t, feature: str, /
    ) -> Sequence[Any]:
        """"""
        if feature in track[0].features:
            return tuple(_cll.features[feature] for _cll in track)

        raise ValueError(f"{feature}: Invalid feature(s)")

    def FeatureEvolutionsAlongAllTracks(
        self, feature: str, /
    ) -> Dict[int, Tuple[single_track_t, Sequence[Any]]]:
        """"""
        output = {}

        for track in self.tracks.single_tracks_iterator:
            output[track.label] = (
                track,
                sequence_t.FeatureEvolutionAlongTrack(track, feature),
            )

        return output

    def ClearContents(self):
        """
        To free up some memory when all processing has been done
        """
        for frames in self.Frames():
            for frame in frames:
                frame.ClearContents()

    def PrintValidInvalidTrackSummary(self) -> None:
        """"""
        n_invalids = []
        issues_per_type = {}
        for track_type, invalids in self.tracks.invalids.items():
            track_type = track_type.__name__[:-2].replace("_", " ").capitalize()
            if (invalids is None) or (invalids.__len__() == 0):
                number = f"{track_type}: None"
            else:
                number = f"{track_type}: {invalids.__len__()}"

                issues = (
                    f"    Track {_tck.labels_as_str}{isse.ISSUE_SEPARATOR}"
                    f"{isse.FactorizedIssuesAsStr(_iss, max_n_contexts=5)}"
                    for _tck, _iss in invalids
                )
                issues_per_type[track_type] = "\n".join(issues)

            n_invalids.append(number)

        n_invalids = ", ".join(n_invalids)
        issues_per_type = "\n".join(
            f"{_typ}:\n{_iss}" for _typ, _iss in issues_per_type.items()
        )

        print(
            f"Tracks: valid={self.tracks.__len__()}; invalid={n_invalids}\n{issues_per_type}"
        )

    def __str__(self) -> str:
        """"""
        if self.tracks is None:
            invalid_tracks = None
        else:
            invalid_tracks = (
                f"{self.tracks.invalids[unstructured_track_t]=}\n"
                f"    {self.tracks.invalids[single_track_t]=}\n"
                f"    {self.tracks.invalids[forking_track_t]=}"
            )

        all_extrema = []
        for channel in self.channels:
            extrema = self.ChannelExtrema(channel)
            all_extrema.append(f"{channel}=[{extrema[0]},{extrema[1]}]")
        all_extrema = "\n    ".join(all_extrema)

        return (
            f"{self.__class__.__name__.upper()}.{ShortID(id(self))}:\n"
            f"    {self.path=}\n"
            f"    {self.base_channels=}\n"
            f"    {self.channels=}\n"
            f"    {all_extrema}\n"
            f"    {self.shape=}\n"
            f"    {self.original_length=}\n"
            f"    {self.length=}\n"
            f"    {self.tracks=}\n"
            f"    {invalid_tracks=}\n"
        )


sequence_h = Union[
    Sequence[array_t], Sequence[segmentation_t], segmentations_t, sequence_t
]


def BoundingBoxSlices(sequence: array_t, /) -> Sequence[slice]:
    """
    sequence: as an XYT-volume
    """
    min_reduction = nmpy.amin(sequence, axis=-1)
    max_reduction = nmpy.amax(sequence, axis=-1)
    constant = nmpy.any(min_reduction != max_reduction, axis=-1)
    output = imge.find_objects(constant)[0]

    return output


def AllChannelsOfSequence(
    sequence: Union[Sequence[array_t], sequence_t]
) -> Tuple[all_versions_h, str]:
    """"""
    if isinstance(sequence, sequence_t):
        all_channels = {}
        for channel in sequence.channels:
            frames = sequence.Frames(channel=channel)
            min_value = min(nmpy.amin(_frm) for _frm in frames)
            max_value = max(nmpy.amax(_frm) for _frm in frames)
            all_channels[channel] = ((min_value, max_value), frames)
        current_channel = sequence.channels[0]
    else:
        current_channel = "MAIN"
        min_value = min(nmpy.amin(_frm) for _frm in sequence)
        max_value = max(nmpy.amax(_frm) for _frm in sequence)
        all_channels = {current_channel: ((min_value, max_value), sequence)}

    return all_channels, current_channel


def AllSegmentationsOfSequence(sequence: sequence_h) -> Tuple[all_versions_h, str]:
    """"""
    if isinstance(sequence, (segmentations_t, sequence_t)):
        if isinstance(sequence, segmentations_t):
            segmentations = sequence
        else:
            segmentations = sequence.segmentations

        all_versions = {}
        compartments, versions = segmentations.available_versions
        for compartment in compartments:
            for version in versions:
                key = f"{compartment.name}:{version[0]}:{version[1]}"
                frames = segmentations.CompartmentsWithVersion(
                    compartment, index=version[0], name=version[1]
                )
                all_versions[key] = ((0, 1), frames)
        current_version = f"{compartment_t.CELL.name}:{versions[0][0]}:{versions[0][1]}"
    else:
        current_version = "MAIN"
        all_versions = {current_version: ((0, 1), sequence)}

    return all_versions, current_version


def AllStreamsOfSequence(sequence: sequence_t) -> Tuple[all_versions_h, str]:
    """"""
    all_streams, current_stream = AllChannelsOfSequence(sequence)
    all_versions, _ = AllSegmentationsOfSequence(sequence)

    all_streams.update(all_versions)

    return all_streams, current_stream
